# -*- coding: utf-8 -*-

# load in basic modules
import os
import string
import numpy as np
from math import pi
from psychopy import data, visual, event
from pylsl.pylsl import StreamInlet, resolve_byprop
from .utils import NeuroScanPort, _check_array_like
import threading
from copy import copy

# prefunctions


def sinusoidal_sample(freqs, phases, srate, frames, stim_color):
    """Sinusoidal approximate sampling method.
    -author: Qiaoyi Wu
    -Created on: 2022-06-20
    -update log:
        2022-06-26 by Jianhang Wu
        2022-08-10 by Wei Zhao
    Parameters
    ----------
        freqs: list of float,
            Frequencies of each stimulus.
        phases: list of float,
            Phases of each stimulus.
        srate: int or float,
            Refresh rate of screen.
        frames: int,
            Flashing frames.
        stim_color: list,
            Color of stimu.
    Returns:
    ----------
        color: ndarray,
            (n_frames, n_elements, 3)
    """

    time = np.linspace(0, (frames - 1) / srate, frames)
    color = np.zeros((frames, len(freqs), 3))
    for ne, (freq, phase) in enumerate(zip(freqs, phases)):
        sinw = np.sin(2 * pi * freq * time + pi * phase) + 1
        color[:, ne, :] = np.vstack(
            (sinw * stim_color[0], sinw * stim_color[1], sinw * stim_color[2])
        ).T
        if stim_color == [-1, -1, -1]:
            pass
        else:
            if stim_color[0] == -1:
                color[:, ne, 0] = -1
            if stim_color[1] == -1:
                color[:, ne, 1] = -1
            if stim_color[2] == -1:
                color[:, ne, 2] = -1

    return color


# create interface for VEP-BCI-Speller


class KeyboardInterface(object):
    """Create stimulus interface.
    -author: Qiaoyi Wu
    -Created on: 2022-06-20
    -update log:
        2022-06-26 by Jianhang Wu
        2022-08-10 by Wei Zhao
    Parameters
    ----------
    win:
        The window object.
    colorspace: str,
        The color space, default to rgb.
    allowGUI: bool
        Defaults to True, which allows frame-by-frame drawing and key-exit.
    """

    def __init__(self, win, colorSpace="rgb", allowGUI=True):
        self.win = win
        win.colorSpace = colorSpace
        win.allowGUI = allowGUI
        win_size = win.size
        self.win_size = np.array(win_size)  # e.g. [1920,1080]

    def config_pos(
        self,
        n_elements=40,
        rows=5,
        columns=8,
        stim_pos=None,
        stim_length=150,
        stim_width=150,
    ):
        """Config positions of stimuli.
        -update log:
            2022-06-26 by Jianhang Wu
        Parameters
        ----------
            n_elements: int,
                Number of stimuli.
            rows: int, optional,
                Rows of keyboard.
            columns: int, optional,
                Columns of keyboard.
            stim_pos: ndarray, optional,
                Extra position matrix.
            stim_length: int,
                Length of stimulus.
            stim_width: int,
                Width of stimulus.
        Raises
        ----------
            Exception: Inconsistent numbers of stimuli and positions.
        """

        self.stim_length = stim_length
        self.stim_width = stim_width
        self.n_elements = n_elements
        # highly customizable position matrix
        if (stim_pos is not None) and (self.n_elements == stim_pos.shape[0]):
            # note that the origin point of the coordinate axis should be the center of your screen
            # (so the upper left corner is in Quadrant 2nd), and the larger the coordinate value,
            # the farther the actual position is from the center
            self.stim_pos = stim_pos
        # conventional design method
        elif (stim_pos is None) and (rows * columns >= self.n_elements):
            # according to the given rows of columns, coordinates will be automatically converted
            stim_pos = np.zeros((self.n_elements, 2))
            # divide the whole screen into rows*columns' blocks, and pick the center of each block
            first_pos = (
                np.array([self.win_size[0] / columns, self.win_size[1] / rows]) / 2
            )
            if (first_pos[0] < stim_length / 2) or (first_pos[1] < stim_width / 2):
                raise Exception("Too much blocks or too big the stimulus region!")
            for i in range(columns):
                for j in range(rows):
                    stim_pos[i * rows + j] = first_pos + [i, j] * first_pos * 2
            # note that those coordinates are still not the real ones that need to be set on the screen
            stim_pos -= self.win_size / 2  # from Quadrant 1st to 3rd
            stim_pos[:, 1] *= -1  # invert the y-axis
            self.stim_pos = stim_pos
        else:
            raise Exception("Incorrect number of stimulus!")

        # check size of stimuli
        stim_sizes = np.zeros((self.n_elements, 2))
        stim_sizes[:] = np.array([stim_length, stim_width])
        self.stim_sizes = stim_sizes
        self.stim_width = stim_width

    def config_text(self, symbols=None, symbol_height=0, tex_color=[1, 1, 1]):
        """Config text stimuli.
        -update log:
            2022-06-26 by Jianhang Wu
        Parameters
        ----------
            symbols: list of str,
                Target characters.
            symbol_height: int,
                Height of target symbol.
            tex_color: list,
                Color of target symbol.
        Raises:
            Exception: Insufficient characters.
        """

        # check number of symbols
        if (symbols is not None) and (len(symbols) >= self.n_elements):
            self.symbols = symbols
        elif self.n_elements <= 40:
            self.symbols = "".join([string.ascii_uppercase, "1234567890+-*/"])
        else:
            raise Exception("Please input correct symbol list!")

        # add text targets onto interface
        if symbol_height == 0:
            symbol_height = self.stim_width / 2
        self.text_stimuli = []
        for symbol, pos in zip(self.symbols, self.stim_pos):
            self.text_stimuli.append(
                visual.TextStim(
                    win=self.win,
                    text=symbol,
                    font="Times New Roman",
                    pos=pos,
                    color=tex_color,
                    units="pix",
                    height=symbol_height,
                    bold=True,
                    name=symbol,
                )
            )

    def config_response(
        self,
        symbol_text="Speller:  ",
        symbol_height=0,
        symbol_color=(1, 1, 1),
        bg_color=[-1, -1, -1],
    ):
        """Config response stimuli.
        -update log:
            2022-08-10 by Wei Zhao
        Parameters
        ----------
            symbol_text: list of str,
                Online response string.
            symbol_height: int,
                Height of response symbol.
            symbol_color: list,
                Color of response symbol.
            bg_color: list,
                Color of background symbol.
        Raises:
            Exception: Insufficient characters.
        """

        brige_length = self.win_size[0] / 2 + self.stim_pos[0][0] - self.stim_length / 2
        brige_width = self.win_size[1] / 2 - self.stim_pos[0][1] - self.stim_width / 2

        self.rect_response = visual.Rect(
            win=self.win,
            units="pix",
            width=self.win_size[0] - brige_length,
            height=brige_width * 3 / 4,
            pos=(0, self.win_size[1] / 2 - brige_width / 2),
            fillColor=bg_color,
            lineColor=[1, 1, 1],
        )

        self.res_text_pos = (
            -self.win_size[0] / 2 + brige_length * 3 / 2,
            self.win_size[1] / 2 - brige_width / 2,
        )
        self.reset_res_pos = (
            -self.win_size[0] / 2 + brige_length * 3 / 2,
            self.win_size[1] / 2 - brige_width / 2,
        )
        self.reset_res_text = "Speller:  "
        if symbol_height == 0:
            self.symbol_height = brige_width / 2
        self.symbol_text = symbol_text
        self.text_response = visual.TextStim(
            win=self.win,
            text=symbol_text,
            font="Times New Roman",
            pos=self.res_text_pos,
            color=symbol_color,
            units="pix",
            height=self.symbol_height,
            bold=True,
        )


# config visual stimuli


class VisualStim(KeyboardInterface):
    """Create various visual stimuli.
    -author: Qiaoyi Wu
    -Created on: 2022-06-20
    -update log:
        2022-06-26 by Jianhang Wu
    Parameters
    ----------
    win:
        The window object.
    colorspace: str,
        The color space, default to rgb.
    allowGUI: bool
        Defaults to True, which allows frame-by-frame drawing and key-exit.
    """

    def __init__(self, win, colorSpace="rgb", allowGUI=True):
        super().__init__(win=win, colorSpace=colorSpace, allowGUI=allowGUI)
        self._exit = threading.Event()

    def config_index(self, index_height=0):
        """Config index stimuli: downward triangle (Unicode: \u2BC6)
        Parameters
        ----------
            index_height: int, optional,
                Defaults to 75 pixels.
        """

        # add index onto interface, with positions to be confirmed.
        if index_height == 0:
            index_height = copy(self.stim_width / 3 * 2)
        self.index_stimuli = visual.TextStim(
            win=self.win,
            text="\u2BC6",
            font="Arial",
            color=[1.0, 1.0, 0.0],
            colorSpace="rgb",
            units="pix",
            height=index_height,
            bold=True,
            autoLog=False,
        )


# standard SSVEP paradigm


class SSVEP(VisualStim):
    """Create SSVEP stimuli.
    -author: Qiaoyi Wu
    -Created on: 2022-06-20
    -update log:
        2022-06-26 by Jianhang Wu
        2022-08-10 by Wei Zhao
    Parameters
    ----------
    win:
        The window object.
    colorspace: str,
        The color space, default to rgb.
    allowGUI: bool
        Defaults to True, which allows frame-by-frame drawing and key-exit.
    """

    def __init__(self, win, colorSpace="rgb", allowGUI=True):
        """Item class from VisualStim.

        Args:

        """
        super().__init__(win=win, colorSpace=colorSpace, allowGUI=allowGUI)

    def config_color(
        self,
        refresh_rate,
        stim_time,
        stim_color,
        stimtype="sinusoid",
        stim_opacities=1,
        **kwargs
    ):
        """Config color of stimuli.
        Parameters
        ----------
            refresh_rate: int or float,
                Refresh rate of screen.
            stim_time: float,
                Time of each stimulus.
            stim_frames: int,
                Flash frames of one trial.
            stim_colors: ndarray,
                (n_frames, n_elements, 3).
            stim_opacities: int or float,
                Opacities of each stimulus.
            freqs: list of float,
                Frequencies of each stimulus.
            phases: list of float,
                Phases of each stimulus.
        Raises:
            Exception: Inconsistent frames and color matrices.
        """

        # initialize extra inputs
        self.refresh_rate = refresh_rate
        self.stim_time = stim_time
        self.stim_color = stim_color
        self.stim_opacities = stim_opacities
        self.stim_frames = int(stim_time * self.refresh_rate)

        if refresh_rate == 0:
            self.refresh_rate = np.floor(
                self.win.getActualFrameRate(nIdentical=20, nWarmUpFrames=20)
            )

        self.stim_oris = np.zeros((self.n_elements,))  # orientation
        self.stim_sfs = np.zeros((self.n_elements,))  # spatial frequency
        self.stim_contrs = np.ones((self.n_elements,))  # contrast

        # check extra inputs
        if "stim_oris" in kwargs.keys():
            self.stim_oris = kwargs["stim_oris"]
        if "stim_sfs" in kwargs.keys():
            self.stim_sfs = kwargs["stim_sfs"]
        if "stim_contrs" in kwargs.keys():
            self.stim_contrs = kwargs["stim_contrs"]
        if "freqs" in kwargs.keys():
            self.freqs = kwargs["freqs"]
        if "phases" in kwargs.keys():
            self.phases = kwargs["phases"]

        # check consistency
        if stimtype == "sinusoid":
            self.stim_colors = sinusoidal_sample(
                freqs=self.freqs,
                phases=self.phases,
                srate=self.refresh_rate,
                frames=self.stim_frames,
                stim_color=stim_color,
            )
            if self.stim_colors[0].shape[0] != self.n_elements:
                raise Exception("Please input correct num of stims!")

        incorrect_frame = self.stim_colors.shape[0] != self.stim_frames
        incorrect_number = self.stim_colors.shape[1] != self.n_elements
        if incorrect_frame or incorrect_number:
            raise Exception("Incorrect color matrix or flash frames!")

        # add flashing targets onto interface
        self.flash_stimuli = []
        for sf in range(self.stim_frames):
            self.flash_stimuli.append(
                visual.ElementArrayStim(
                    win=self.win,
                    units="pix",
                    nElements=self.n_elements,
                    sizes=self.stim_sizes,
                    xys=self.stim_pos,
                    colors=self.stim_colors[sf, ...],
                    opacities=self.stim_opacities,
                    oris=self.stim_oris,
                    sfs=self.stim_sfs,
                    contrs=self.stim_contrs,
                    elementTex=np.ones((64, 64)),
                    elementMask=None,
                    texRes=48,
                )
            )


# standard P300 paradigm


class P300(VisualStim):
    """Create P300 stimuli.
    -author: Shengfu Wen
    -Created on: 2022-07-04
    -update log:
        2022-08-10 by Wei Zhao
    Parameters
    ----------
    win:
        The window object.
    colorspace: str,
        The color space, default to rgb.
    allowGUI: bool
        Defaults to True, which allows frame-by-frame drawing and key-exit.
    """

    def __init__(self, win, colorSpace="rgb", allowGUI=True):
        super().__init__(win=win, colorSpace=colorSpace, allowGUI=allowGUI)

    def config_color(self, refresh_rate=0, stim_duration=0.5):
        """Config color of stimuli.
        Parameters
        ----------
            refresh_rate: int or float,
                Refresh rate of screen.
            symbol_height: float,
                Height of each stimulus.
            stim_duration: float,
                The duration of one trial.
        """
        self.stim_duration = stim_duration
        self.refresh_rate = refresh_rate
        if refresh_rate == 0:
            self.refresh_rate = np.floor(
                self.win.getActualFrameRate(nIdentical=20, nWarmUpFrames=20)
            )

        # highlight one row/ column onto interface
        row_pos = np.unique(self.stim_pos[:, 0])
        col_pos = np.unique(self.stim_pos[:, 1])
        [row_num, col_num] = [len(col_pos), len(row_pos)]
        # complete single trial
        self.stim_frames = int((row_num + col_num) * stim_duration * refresh_rate)

        row_order_index = list(range(0, row_num))
        np.random.shuffle(row_order_index)
        col_order_index = list(range(0, col_num))
        np.random.shuffle(col_order_index)
        l_row_order_index = [
            x + self.n_elements + 1 for x in row_order_index
        ]  # reset event label
        l_col_order_index = [x + self.n_elements + row_num + 1 for x in col_order_index]

        self.order_index = np.array(
            l_row_order_index + l_col_order_index
        )  # event label

        # Determine row and column char status
        stim_colors_row = np.zeros(
            [(row_num * col_num), int(row_num * refresh_rate * stim_duration), 3]
        )
        stim_colors_col = np.zeros(
            [(row_num * col_num), int(col_num * refresh_rate * stim_duration), 3]
        )  #

        tmp = 0
        for col_i in col_order_index:
            stim_colors_col[
                (col_i * row_num) : ((col_i + 1) * row_num),
                int(tmp * refresh_rate * stim_duration) : int(
                    (tmp + 1) * refresh_rate * stim_duration
                ),
            ] = [-1, -1, -1]
            tmp += 1

        tmp = 0
        for row_i in row_order_index:
            for col_i in range(col_num):
                stim_colors_row[
                    (row_i + row_num * col_i),
                    int(tmp * refresh_rate * stim_duration) : int(
                        (tmp + 1) * refresh_rate * stim_duration
                    ),
                ] = [-1, -1, -1]
            tmp += 1

        stim_colors = np.concatenate((stim_colors_row, stim_colors_col), axis=1)
        self.stim_colors = np.transpose(stim_colors, [1, 0, 2])

        # add flashing targets onto interface
        self.flash_stimuli = []
        for sf in range(self.stim_frames):
            self.flash_stimuli.append(
                visual.ElementArrayStim(
                    win=self.win,
                    units="pix",
                    nElements=self.n_elements,
                    sizes=self.stim_sizes,
                    xys=self.stim_pos,
                    colors=self.stim_colors[sf, ...],
                    elementTex=np.ones((64, 64)),
                    elementMask=None,
                    texRes=48,
                )
            )


# standard MI paradigm


class MI(VisualStim):
    """Create MI stimuli.
    -author: Wei Zhao
    -Created on: 2022-06-30
    -update log:
        2022-08-10 by Wei Zhao
    Parameters
    ----------
    win:
        The window object.
    colorspace: str,
        The color space, default to rgb.
    allowGUI: bool
        Defaults to True, which allows frame-by-frame drawing and key-exit.
    """

    def __init__(self, win, colorSpace="rgb", allowGUI=True):
        super().__init__(win=win, colorSpace=colorSpace, allowGUI=allowGUI)

        self.tex_left = os.path.join(
            os.path.abspath(os.path.dirname(os.path.abspath(__file__))),
            "textures\\left_hand.png",
        )
        self.tex_right = os.path.join(
            os.path.abspath(os.path.dirname(os.path.abspath(__file__))),
            "textures\\right_hand.png",
        )

    def config_color(
        self,
        refresh_rate=60,
        text_pos=(0.0, 0.0),
        left_pos=[[-480, 0.0]],
        right_pos=[[480, 0.0]],
        tex_color=(1, -1, -1),
        normal_color=[[-0.8, -0.8, 0.8]],
        image_color=[[1, 1, 1]],
        symbol_height=100,
        n_Elements=1,
        stim_length=288,
        stim_width=162,
    ):
        """Config color of stimuli.
        Parameters
        ----------
            refresh_rate: int or float,
                Refresh rate of screen.
            text_pos: ndarray,
                Position of text.
            left_pos: list,
                Position of left hand.
            right_pos: list,
                Position of right hand.
            text_color: list,
                Color of text.
            normal_color: list,
                Color of default stimulus
            image_color: list,
                Color of image or indicate stimulus.
            symbol_height: list,
                Height of text.
            n_Elements: list,
                Num of stimulus.
            stim_length: list,
                Length of stimulus.
            stim_width: list,
                Width of stimulus.
        """

        self.n_Elements = n_Elements
        self.stim_length = stim_length
        self.stim_width = stim_width
        self.left_pos = left_pos
        self.right_pos = right_pos
        self.refresh_rate = refresh_rate
        if refresh_rate == 0:
            refresh_rate = np.floor(
                self.win.getActualFrameRate(nIdentical=20, nWarmUpFrames=20)
            )

        if symbol_height == 0:
            symbol_height = int(self.win_size[1] / 6)
        self.text_stimulus = visual.TextStim(
            self.win,
            text="start",
            font="Times New Roman",
            pos=text_pos,
            color=tex_color,
            units="pix",
            height=symbol_height,
            bold=True,
        )

        self.image_left_stimuli = visual.ElementArrayStim(
            self.win,
            units="pix",
            elementTex=self.tex_left,
            elementMask=None,
            texRes=2,
            nElements=n_Elements,
            sizes=[[stim_length, stim_width]],
            xys=np.array(left_pos),
            oris=[0],
            colors=np.array(image_color),
            opacities=[1],
            contrs=[-1],
        )
        self.image_right_stimuli = visual.ElementArrayStim(
            self.win,
            units="pix",
            elementTex=self.tex_right,
            elementMask=None,
            texRes=2,
            nElements=n_Elements,
            sizes=[[stim_length, stim_width]],
            xys=np.array(right_pos),
            oris=[0],
            colors=np.array(image_color),
            opacities=[1],
            contrs=[-1],
        )

        self.normal_left_stimuli = visual.ElementArrayStim(
            self.win,
            units="pix",
            elementTex=self.tex_left,
            elementMask=None,
            texRes=2,
            nElements=n_Elements,
            sizes=[[stim_length, stim_width]],
            xys=np.array(left_pos),
            oris=[0],
            colors=np.array(normal_color),
            opacities=[1],
            contrs=[-1],
        )
        self.normal_right_stimuli = visual.ElementArrayStim(
            self.win,
            units="pix",
            elementTex=self.tex_right,
            elementMask=None,
            texRes=2,
            nElements=n_Elements,
            sizes=[[stim_length, stim_width]],
            xys=np.array(right_pos),
            oris=[0],
            colors=np.array(normal_color),
            opacities=[1],
            contrs=[-1],
        )

    def config_response(self, response_color=[[-0.5, 0.9, 0.5]]):
        self.response_left_stimuli = visual.ElementArrayStim(
            self.win,
            units="pix",
            elementTex=self.tex_left,
            elementMask=None,
            texRes=2,
            nElements=self.n_Elements,
            sizes=[[self.stim_length, self.stim_width]],
            xys=np.array(self.left_pos),
            oris=[0],
            colors=np.array(response_color),
            opacities=[1],
            contrs=[-1],
        )
        self.response_right_stimuli = visual.ElementArrayStim(
            self.win,
            units="pix",
            elementTex=self.tex_right,
            elementMask=None,
            texRes=2,
            nElements=self.n_Elements,
            sizes=[[self.stim_length, self.stim_width]],
            xys=np.array(self.right_pos),
            oris=[0],
            colors=np.array(response_color),
            opacities=[1],
            contrs=[-1],
        )


# continuous experimen


class GetPlabel_MyTherad:
    """Start a thread that receives online results
    -author: Wei Zhao
    -Created on: 2022-07-30
    -update log:
        2022-08-10 by Wei Zhao
    Parameters
    ----------
    inlet:
        Stream data online.
    """

    def __init__(self, inlet):
        self.inlet = inlet
        self._exit = threading.Event()

    def feedbackThread(self):
        """Start the thread."""
        self._t_loop = threading.Thread(
            target=self._inner_loop, name="get_predict_id_loop"
        )
        self._t_loop.start()

    def _inner_loop(self):
        """The inner loop in the thread."""
        self._exit.clear()
        global online_text_pos, online_symbol_text
        online_text_pos = copy(self.res_text_pos)
        online_symbol_text = copy(self.symbol_text)
        while not self._exit.is_set():
            try:
                samples, _ = self.inlet.pull_sample()
                if samples:
                    # online predict id
                    predict_id = int(samples[0]) - 1
                    online_text_pos = (
                        online_text_pos[0] + self.symbol_height / 3,
                        online_text_pos[1],
                    )
                    online_symbol_text = online_symbol_text + self.symbols[predict_id]
            except Exception:
                pass

    def stop_feedbackThread(self):
        """Stop the thread."""
        self._exit.set()
        self._t_loop.join()


# basic experiment control


def paradigm(
    VSObject,
    win,
    bg_color,
    display_time=1.0,
    index_time=1.0,
    rest_time=0.5,
    response_time=2,
    image_time=2,
    port_addr=9045,
    nrep=1,
    pdim="ssvep",
    lsl_source_id=None,
    online=None,
):
    """Passing outsied parameters to inner attributes.
    -author: Wei Zhao
    -Created on: 2022-07-30
    -update log:
        2022-08-10 by Wei Zhao
        2022-08-03 by Shengfu Wen
    Parameters
    ----------
        bg_color: ndarray,
            Background color.
        display_time: float,
            Keyboard display time before 1st index.
        index_time: float,
            Indicator display time.
        rest_time: float, optional,
            Rest-state time.
        respond_time: float, optional,
            Feedback time during online experiment.
        image_time: float, optional,
            Image time.
        port_addr:
             Computer port.
        nrep: int,
            Num of blocks.
        mi_flag: bool,
            Flag of MI paradigm.
        lsl_source_id: str,
            Source id.
        online: bool,
            Flag of online experiment.
    """

    if not _check_array_like(bg_color, 3):
        raise ValueError("bg_color should be 3 elements array-like object.")
    win.color = bg_color
    fps = VSObject.refresh_rate

    port = NeuroScanPort(port_addr, use_serial=False) if port_addr else None
    port_frame = int(0.05 * fps)

    inlet = False
    if online:
        if pdim == "ssvep" or pdim == "p300" or pdim == "con-ssvep":
            VSObject.text_response.text = copy(VSObject.reset_res_text)
            VSObject.text_response.pos = copy(VSObject.reset_res_pos)
            VSObject.res_text_pos = copy(VSObject.reset_res_pos)
            VSObject.symbol_text = copy(VSObject.reset_res_text)
            res_text_pos = VSObject.reset_res_pos
        if lsl_source_id:
            inlet = True
            streams = resolve_byprop(
                "source_id", lsl_source_id, timeout=5
            )  # Resolve all streams by source_id
            if not streams:
                return
            inlet = StreamInlet(streams[0])  # receive stream data

    if pdim == "ssvep":
        # config experiment settings
        conditions = [{"id": i} for i in range(VSObject.n_elements)]
        trials = data.TrialHandler(conditions, nrep, name="experiment", method="random")

        # start routine
        # episode 1: display speller interface
        iframe = 0
        while iframe < int(fps * display_time):
            if online:
                VSObject.rect_response.draw()
                VSObject.text_response.draw()
            for text_stimulus in VSObject.text_stimuli:
                text_stimulus.draw()
            iframe += 1
            win.flip()

        # episode 2: begin to flash
        if port:
            port.setData(0)
        for trial in trials:
            # quit demo
            keys = event.getKeys(["q"])
            if "q" in keys:
                break

            # initialise index position
            id = int(trial["id"])
            position = VSObject.stim_pos[id] + np.array([0, VSObject.stim_width / 2])
            VSObject.index_stimuli.setPos(position)

            # phase I: speller & index (eye shifting)
            iframe = 0
            while iframe < int(fps * index_time):
                if online:
                    VSObject.rect_response.draw()
                    VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                VSObject.index_stimuli.draw()
                iframe += 1
                win.flip()

            # phase II: rest state
            if rest_time != 0:
                iframe = 0
                while iframe < int(fps * rest_time):
                    if online:
                        VSObject.rect_response.draw()
                        VSObject.text_response.draw()
                    for text_stimulus in VSObject.text_stimuli:
                        text_stimulus.draw()
                    iframe += 1
                    win.flip()

            # phase III: target stimulating
            for sf in range(VSObject.stim_frames):
                if sf == 0 and port and online:
                    VSObject.win.callOnFlip(port.setData, 1)
                elif sf == 0 and port:
                    VSObject.win.callOnFlip(port.setData, id + 1)
                if sf == port_frame and port:
                    port.setData(0)
                VSObject.flash_stimuli[sf].draw()
                if online:
                    VSObject.rect_response.draw()
                    VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                win.flip()

            # phase IV: respond
            if inlet:
                VSObject.rect_response.draw()
                VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                win.flip()
                samples, timestamp = inlet.pull_sample()
                predict_id = int(samples[0]) - 1  # online predict id
                VSObject.symbol_text = (
                    VSObject.symbol_text + VSObject.symbols[predict_id]
                )
                res_text_pos = (
                    res_text_pos[0] + VSObject.symbol_height / 3,
                    res_text_pos[1],
                )
                iframe = 0
                while iframe < int(fps * response_time):
                    for text_stimulus in VSObject.text_stimuli:
                        text_stimulus.draw()
                    VSObject.rect_response.draw()
                    VSObject.text_response.text = VSObject.symbol_text
                    VSObject.text_response.pos = res_text_pos
                    VSObject.text_response.draw()
                    iframe += 1
                    win.flip()

    elif pdim == "p300":
        # config experiment settings
        conditions = [{"id": i} for i in range(VSObject.n_elements)]
        trials = data.TrialHandler(conditions, nrep, name="experiment", method="random")

        # start routine
        # episode 1: display speller interface
        iframe = 0
        while iframe < int(fps * display_time):
            if online:
                VSObject.rect_response.draw()
                VSObject.text_response.draw()
            for text_stimulus in VSObject.text_stimuli:
                text_stimulus.draw()
            iframe += 1
            win.flip()

        # episode 2: begin to flash
        if port:
            port.setData(0)
        for trial in trials:
            # quit demo
            keys = event.getKeys(["q"])
            if "q" in keys:
                break

            # initialise index position
            id = int(trial["id"])
            position = VSObject.stim_pos[id] + np.array([0, VSObject.stim_width / 2])
            VSObject.index_stimuli.setPos(position)

            # phase I: speller & index (eye shifting)
            iframe = 0
            while iframe < int(fps * index_time):
                if iframe == 0 and port and online:
                    VSObject.win.callOnFlip(port.setData, 1)
                elif iframe == 0 and port:
                    VSObject.win.callOnFlip(port.setData, id + 1)
                if iframe == port_frame and port:
                    port.setData(0)
                if online:
                    VSObject.rect_response.draw()
                    VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                VSObject.index_stimuli.draw()
                iframe += 1
                win.flip()

            # phase II: rest state
            if rest_time != 0:
                iframe = 0
                while iframe < int(fps * rest_time):
                    if online:
                        VSObject.rect_response.draw()
                        VSObject.text_response.draw()
                    for text_stimulus in VSObject.text_stimuli:
                        text_stimulus.draw()
                    iframe += 1
                    win.flip()

            # phase III: target stimulating
            for sf in range(VSObject.stim_frames):
                if (sf % (VSObject.stim_duration * fps)) == 0 and port:
                    VSObject.win.callOnFlip(
                        port.setData,
                        VSObject.order_index[int(sf / (VSObject.stim_duration * fps))],
                    )
                if (sf % (VSObject.stim_duration * fps)) == port_frame and port:
                    port.setData(0)

                VSObject.flash_stimuli[sf].draw()
                if online:
                    VSObject.rect_response.draw()
                    VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                win.flip()

            # phase IV: respond
            if inlet:
                VSObject.rect_response.draw()
                VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                win.flip()
                samples, timestamp = inlet.pull_sample()
                predict_id = int(samples[0]) - 1  # online predict id
                VSObject.symbol_text = (
                    VSObject.symbol_text + VSObject.symbols[predict_id]
                )
                res_text_pos = (
                    res_text_pos[0] + VSObject.symbol_height / 3,
                    res_text_pos[1],
                )
                iframe = 0
                while iframe < int(fps * response_time):
                    for text_stimulus in VSObject.text_stimuli:
                        text_stimulus.draw()
                    VSObject.rect_response.draw()
                    VSObject.text_response.text = VSObject.symbol_text
                    VSObject.text_response.pos = res_text_pos
                    VSObject.text_response.draw()
                    iframe += 1
                    win.flip()

    elif pdim == "mi":
        # config experiment settings
        conditions = [
            {"id": 0, "name": "left_hand"},
            {"id": 1, "name": "right_hand"},
            {"id": 2, "name": "both_hands"},
        ]
        trials = data.TrialHandler(conditions, nrep, name="experiment", method="random")

        # start routine
        # episode 1: display speller interface
        iframe = 0
        while iframe < int(fps * display_time):
            VSObject.normal_left_stimuli.draw()
            VSObject.normal_right_stimuli.draw()
            iframe += 1
            win.flip()

        # episode 2: begin to flash
        if port:
            port.setData(0)
        for trial in trials:
            # quit demo
            keys = event.getKeys(["q"])
            if "q" in keys:
                break

            # initialise index position
            id = int(trial["id"])
            if id == 0:
                image_stimuli = [VSObject.image_left_stimuli]
                normal_stimuli = [VSObject.normal_right_stimuli]
            elif id == 1:
                image_stimuli = [VSObject.image_right_stimuli]
                normal_stimuli = [VSObject.normal_left_stimuli]
            else:
                image_stimuli = [
                    VSObject.image_left_stimuli,
                    VSObject.image_right_stimuli,
                ]
                normal_stimuli = []

            # phase I: rest state
            if rest_time != 0:
                iframe = 0
                while iframe < int(fps * rest_time):
                    VSObject.normal_left_stimuli.draw()
                    VSObject.normal_right_stimuli.draw()
                    iframe += 1
                    win.flip()

            # phase II: speller & index (eye shifting)
            iframe = 0
            while iframe < int(fps * index_time):
                for _image_stimuli in image_stimuli:
                    _image_stimuli.draw()
                if normal_stimuli:
                    for _normal_stimuli in normal_stimuli:
                        _normal_stimuli.draw()
                iframe += 1
                win.flip()

            # phase III: target stimulating
            iframe = 0
            while iframe < int(fps * image_time):
                if iframe == 0 and port and online:
                    VSObject.win.callOnFlip(port.setData, 1)
                elif iframe == 0 and port:
                    VSObject.win.callOnFlip(port.setData, id + 1)
                if iframe == port_frame and port:
                    port.setData(0)
                VSObject.text_stimulus.draw()
                for _image_stimuli in image_stimuli:
                    _image_stimuli.draw()
                if normal_stimuli:
                    for _normal_stimuli in normal_stimuli:
                        _normal_stimuli.draw()
                iframe += 1
                win.flip()

            # phase IV: respond
            if inlet:
                VSObject.normal_left_stimuli.draw()
                VSObject.normal_right_stimuli.draw()
                win.flip()

                samples, timestamp = inlet.pull_sample()
                predict_id = int(samples[0]) - 1  # online predict id

                if predict_id == 0:
                    response_stimuli = [VSObject.response_left_stimuli]
                    normal_stimuli = [VSObject.normal_right_stimuli]
                elif predict_id == 1:
                    response_stimuli = [VSObject.response_right_stimuli]
                    normal_stimuli = [VSObject.normal_left_stimuli]
                else:
                    response_stimuli = [
                        VSObject.response_left_stimuli,
                        VSObject.response_right_stimuli,
                    ]
                    normal_stimuli = []

                iframe = 0
                while iframe < int(fps * response_time):
                    for _response_stimuli in response_stimuli:
                        _response_stimuli.draw()
                    if normal_stimuli:
                        for _normal_stimuli in normal_stimuli:
                            _normal_stimuli.draw()
                    iframe += 1
                    win.flip()

    elif pdim == "con-ssvep":
        global online_text_pos, online_symbol_text

        if inlet:
            MyTherad = GetPlabel_MyTherad(inlet)
            MyTherad.feedbackThread()

        # config experiment settings
        conditions = [{"id": i} for i in range(VSObject.n_elements)]
        trials = data.TrialHandler(conditions, nrep, name="experiment", method="random")

        # start routine
        # episode 1: display speller interface
        iframe = 0
        while iframe < int(fps * display_time):
            if online:
                VSObject.rect_response.draw()
                VSObject.text_response.draw()
            for text_stimulus in VSObject.text_stimuli:
                text_stimulus.draw()
            iframe += 1
            win.flip()

        # episode 2: begin to flash
        if port:
            port.setData(0)
        for trial in trials:
            # quit demo
            keys = event.getKeys(["q"])
            if "q" in keys:
                MyTherad.stop_feedbackThread()
                break

            # initialise index position
            id = int(trial["id"])
            position = VSObject.stim_pos[id] + np.array([0, VSObject.stim_width / 2])
            VSObject.index_stimuli.setPos(position)

            # phase I: speller & index (eye shifting)
            iframe = 0
            while iframe < int(fps * index_time):
                if online:
                    VSObject.rect_response.draw()
                    VSObject.text_response.text = online_symbol_text
                    VSObject.text_response.pos = online_text_pos
                    VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                VSObject.index_stimuli.draw()
                iframe += 1
                win.flip()

            # phase II: rest state
            if rest_time != 0:
                iframe = 0
                while iframe < int(fps * rest_time):
                    if online:
                        VSObject.rect_response.draw()
                        VSObject.text_response.text = online_symbol_text
                        VSObject.text_response.pos = online_text_pos
                        VSObject.text_response.draw()
                    for text_stimulus in VSObject.text_stimuli:
                        text_stimulus.draw()
                    iframe += 1
                    win.flip()

            # phase III: target stimulating
            for sf in range(VSObject.stim_frames):
                if sf == 0 and port and online:
                    VSObject.win.callOnFlip(port.setData, 1)
                elif sf == 0 and port:
                    VSObject.win.callOnFlip(port.setData, id + 1)
                if sf == port_frame and port:
                    port.setData(0)
                VSObject.flash_stimuli[sf].draw()
                if online:
                    VSObject.rect_response.draw()
                    VSObject.text_response.text = online_symbol_text
                    VSObject.text_response.pos = online_text_pos
                    VSObject.text_response.draw()
                for text_stimulus in VSObject.text_stimuli:
                    text_stimulus.draw()
                win.flip()

        if inlet:
            MyTherad.stop_feedbackThread()



# -*- coding: utf-8 -*-
"""
China BCI Competition.
"""
from metabci.brainda.utils.download import mne_data_path
from typing import Union, Optional, Dict, List, cast
from pathlib import Path

import numpy as np
from mne import create_info
from mne.io import Raw, RawArray
from mne.channels import make_standard_montage
from .base import BaseDataset
from ..utils.channels import upper_ch_names
from ..utils.io import loadmat

# no available links now
CBCIC2019001_URL = "file:///CBCIC2019001"
CBCIC2019004_URL = "file:///CBCIC2019004"


class CBCIC2019001(BaseDataset):
    """2019 China BCI competition Dataset for MI in preliminary contest A/B.

    Motor imagery dataset from China BCI competition in 2019.

    This dataset contains EEG recordings from 18 subjects, performing 2 or 3 tasks
    of motor imagery (left hand, right hand or feet). Data have been recored at 1000hz
    with 64 electrodes (59 in use except ECG, HEOR, HEOL, VEOU, VEOL channels) by
    an neuracle EEG amplifier.

    """

    _EVENTS = {
        "left_hand": (1, (1.5, 5.5)),
        "right_hand": (2, (1.5, 5.5)),
        "feet": (3, (1.5, 5.5)),
        # "open_eye_relax": (7, (0, 60)),
        # "close_eye_relax": (8, (0, 60))
    }

    _CHANNELS = [
        "FPZ",
        "FP1",
        "FP2",
        "AF3",
        "AF4",
        "AF7",
        "AF8",
        "FZ",
        "F1",
        "F2",
        "F3",
        "F4",
        "F5",
        "F6",
        "F7",
        "F8",
        "FCZ",
        "FC1",
        "FC2",
        "FC3",
        "FC4",
        "FC5",
        "FC6",
        "FT7",
        "FT8",
        "CZ",
        "C1",
        "C2",
        "C3",
        "C4",
        "C5",
        "C6",
        "T7",
        "T8",
        "CP1",
        "CP2",
        "CP3",
        "CP4",
        "CP5",
        "CP6",
        "TP7",
        "TP8",
        "PZ",
        "P3",
        "P4",
        "P5",
        "P6",
        "P7",
        "P8",
        "POZ",
        "PO3",
        "PO4",
        "PO5",
        "PO6",
        "PO7",
        "PO8",
        "OZ",
        "O1",
        "O2",
    ]

    def __init__(self):
        super().__init__(
            dataset_code="cbcic2019001",
            subjects=list(range(1, 19)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=1000,
            paradigm="imagery",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:
        subject = cast(int, subject)
        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        if subject in [6, 14, 15, 18]:
            file_name = "T{:02d}01T.mat".format(subject)
        else:
            file_name = "B{:02d}01T.mat".format(subject)

        url = "{:s}/{:02d}/{:s}".format(CBCIC2019001_URL, subject, file_name)
        dests = [
            [
                mne_data_path(
                    url,
                    "cbcic",
                    path=path,
                    proxies=proxies,
                    force_update=force_update,
                    update_path=update_path,
                )
            ]
        ]
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)
        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]

        sess = dict()
        for isess, run_dests in enumerate(dests):
            runs = dict()
            for irun, run_file in enumerate(run_dests):
                raw_mat = loadmat(run_file)["EEG"]
                epoch_data = raw_mat["data"][:-5] * 1e-6
                stim = np.zeros((1, epoch_data.shape[-1]))
                for event in raw_mat["event"]:
                    stim[0, int(event["latency"]) - 1] = int(event["type"])
                data = np.concatenate((epoch_data, stim), axis=0)

                ch_names = [ch_name.upper() for ch_name in self._CHANNELS]
                ch_types = ["eeg"] * len(ch_names)
                ch_names = ch_names + ["STI 014"]
                ch_types = ch_types + ["stim"]

                info = create_info(
                    ch_names=ch_names, ch_types=ch_types, sfreq=self.srate
                )

                raw = RawArray(data=data, info=info)
                raw = upper_ch_names(raw)
                raw.set_montage(montage)
                runs["run_{:d}".format(irun)] = raw
            sess["session_{:d}".format(isess)] = runs
        return sess


class CBCIC2019004(BaseDataset):
    """2019 China BCI competition Dataset for MI in final competition.

    Motor imagery dataset from China BCI competition in 2019.

    This dataset contains EEG recordings from 18 subjects, performing 2 or 3 tasks
    of motor imagery (left hand, right hand or feet). Data have been recored at 1000hz
    with 64 electrodes (59 in use except ECG, HEOR, HEOL, VEOU, VEOL channels) by
    an neuracle EEG amplifier.

    """

    _EVENTS = {
        "left_hand": (1, (0, 4)),
        "right_hand": (2, (0, 4)),
    }

    _CHANNELS = [
        "FPZ",
        "FP1",
        "FP2",
        "AF3",
        "AF4",
        "AF7",
        "AF8",
        "FZ",
        "F1",
        "F2",
        "F3",
        "F4",
        "F5",
        "F6",
        "F7",
        "F8",
        "FCZ",
        "FC1",
        "FC2",
        "FC3",
        "FC4",
        "FC5",
        "FC6",
        "FT7",
        "FT8",
        "CZ",
        "C1",
        "C2",
        "C3",
        "C4",
        "C5",
        "C6",
        "T7",
        "T8",
        "CP1",
        "CP2",
        "CP3",
        "CP4",
        "CP5",
        "CP6",
        "TP7",
        "TP8",
        "PZ",
        "P3",
        "P4",
        "P5",
        "P6",
        "P7",
        "P8",
        "POZ",
        "PO3",
        "PO4",
        "PO5",
        "PO6",
        "PO7",
        "PO8",
        "OZ",
        "O1",
        "O2",
    ]

    def __init__(self):
        super().__init__(
            dataset_code="cbcic2019004",
            subjects=list(range(1, 7)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=250,
            paradigm="imagery",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:

        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        subject = cast(int, subject)
        runs = []
        for i in range(1, 5):
            url = "{:s}/{:02d}/block{:d}.mat".format(CBCIC2019004_URL, subject, i)
            runs.append(
                mne_data_path(
                    url,
                    "cbcic",
                    path=path,
                    proxies=proxies,
                    force_update=force_update,
                    update_path=update_path,
                )
            )

        dests = [runs]
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)
        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]

        sess = dict()
        for isess, run_dests in enumerate(dests):
            runs = dict()
            for irun, run_file in enumerate(run_dests):
                raw_mat = loadmat(run_file)
                epoch_data = raw_mat["data"][:-6] * 1e-6
                stims = raw_mat["data"][-1][np.newaxis, :]
                data = np.concatenate((epoch_data, stims), axis=0)

                ch_names = [ch_name.upper() for ch_name in self._CHANNELS]
                ch_types = ["eeg"] * len(ch_names)
                ch_names = ch_names + ["STI 014"]
                ch_types = ch_types + ["stim"]

                info = create_info(
                    ch_names=ch_names, ch_types=ch_types, sfreq=self.srate
                )

                raw = RawArray(data=data, info=info)
                raw = upper_ch_names(raw)
                raw.set_montage(montage)
                runs["run_{:d}".format(irun)] = raw
            sess["session_{:d}".format(isess)] = runs
        return sess



import numpy as np
from scipy.signal import detrend, stft
import mne
from scipy.signal import hilbert
import matplotlib.pyplot as plt


class TimeFrequencyAnalysis:
    def __init__(self, fs):
        self.fs = fs

    def func_morlet_wavelet(self, data, xtimes, omega, sigma, fs=None):
        """
        -author: Xin Fengran
        -Created on: 2022-8-8
        -update log:
            2022-8-11 by Xin Fengran
        Args:
            data: ndarray(Nchannel, Ntimes), the EEG data
            xtimes: ndarray(N,), timeline of the EEG data
            omega: float.
            sigma: float.
            fs: float, the resampling rate
        Returns:
            P: ndarray(Nchannel, N, nTimes), square amplitude of Morlet wavelet transform
            S: ndarray(Nchannel, N, nTimes), complex values of Morlet wavelet transform

        """
        data = detrend(data, axis=-1, type="linear")
        N_T = data.shape[1]
        N_C = data.shape[0]
        N_F = xtimes.shape[0]
        if fs is None:
            fs = self.fs
        f = xtimes / fs
        S = np.zeros((N_C, N_F, N_T))
        P = np.zeros((N_C, N_F, N_T))

        L_hw = N_T
        for fi in range(N_F):
            scaling_factor = omega / f[fi]
            u = (-np.arange(-L_hw, L_hw + 1)) / scaling_factor
            hw = (
                np.sqrt(1 / scaling_factor)
                * np.exp(-(u**2) / (2 * sigma**2))
                * np.exp(1j * 2 * np.pi * omega * u)
            )
            for ci in range(N_C):
                S_full = np.convolve(data[ci, :], hw.conjugate())
                S[ci, fi, :] = S_full[L_hw : L_hw + N_T]

        P = np.abs(S) ** 2
        return P, S

    def fun_stft(
        self,
        data,
        fs=None,
        window="hann",
        nperseg=256,
        noverlap=None,
        nfft=None,
        detrend=False,
        return_onesided=True,
        boundary="zeros",
        padded=True,
        axis=-1,
    ):
        """
        -author: Xin Fengran
        -Created on: 2022-8-8
        -updata log:
            2022-8-11 by Xin Fengran
        Args:
            data(ndarray): the EEG data
            fs(float): the rasampling rate
            window(str or tuple or ndarray): desired window to use.
            nperseg(int): length of each segment.
            noverlap(int): number of points to overlap between segments.
            nfft(int): length of the FFT used.
            detrend(str or function or False): specifies how to detrend each segment.
            return_onesided(bool): If True, return a one-sided spectrum for real data. If False return a two-sided spectrum.
            Defaults to True, but for complex data, a two-sided spectrum is always returned.
            boundary(str): Specifies whether the input signal is extended at both ends, and how to generate the new values,
            in order to center the first windowed segment on the first input point.
            padded(bool): Specifies whether the input signal is zero-padded at the end.
            axis(int): axis along which the STFT is computed
        Returns:
            f(ndarray): array of sample frequencies
            t(ndarray): array of segment times
            Zxx(ndarray): the STFT of the EEG data
        """
        if fs is None:
            fs = self.fs
        f, t, Zxx = stft(
            data,
            fs,
            window,
            nperseg,
            noverlap,
            nfft,
            detrend,
            return_onesided,
            boundary,
            padded,
            axis,
        )
        return f, t, Zxx

    def fun_topoplot(self, X, chan_names, sfreq=None, ch_types="eeg"):
        """
        -author: Li Xiaoyu
        -Created on: 2022-8-8
        -updata log:
            2022-8-11 by Li Xiaoyu

        Args:
            X(ndarray): the input data
            chan_names(list): the name of channels
            sfreq(float): the sampling rate
            ch_types(str): the type of channel

        """
        if sfreq is None:
            sfreq = self.fs

        montage_type = mne.channels.make_standard_montage(
            "standard_1020", head_size=0.1
        )
        epoch_info = mne.create_info(
            ch_names=chan_names, ch_types=ch_types, sfreq=sfreq
        )
        epoch_info.set_montage(montage_type)

        # The value specifying the upper bound of the color range.
        vmax = np.max(X)
        # The value specifying the lower bound of the color range.
        vmin = np.min(X)

        # plot
        fig, ax = plt.subplots(1, figsize=(4, 4), sharex=True, sharey=True)
        im, cn = mne.viz.plot_topomap(
            X, epoch_info, axes=ax, show=False, vmax=vmax, vmin=vmin, cnorm=None
        )
        ax.set_title("topomap", fontsize=25)
        plt.colorbar(im)
        plt.show()

    def fun_hilbert(self, X, N=None, axis=-1):
        """
        -author: Li Xiaoyu
        -Created on: 2022-8-8
        -updata log:
            2022-8-11 by Li Xiaoyu

        Args:
            X(ndarray): the input data
            N(int): length of the hilbert used.
        Returns:
            analytic_signal(ndarray): discrete-time analytic signal
            realEnv(ndarray): the real part of the discrete-time analytic signal.
            imagEnv(ndarray): the imaginary part of the discrete-time analytical signal.
            angle(ndarray): the angle of the discrete-time analytical signal.
            envModu(ndarray): the envelope of the discrete-time analytical signal

        """
        analytic_signal = hilbert(X, N, axis)
        realEnv = np.real(analytic_signal)
        imagEnv = np.imag(analytic_signal)
        angle = np.angle(analytic_signal)
        envModu = np.sqrt(realEnv**2 + imagEnv**2)
        return analytic_signal, realEnv, imagEnv, angle, envModu



# -*- coding: utf-8 -*-
# License: MIT License
"""
Many Workers.

"""
from typing import Optional, Any
from abc import abstractmethod
import os
import multiprocessing
import queue
from .logger import get_logger

logger = get_logger("worker")


class ProcessWorker(multiprocessing.Process):
    """Online processing.
    -author: Lichao Xu
    -Created on: 2021-04-01
    -update log:
        2022-08-10 by Wei Zhao
    """

    def __init__(self, timeout: float = 1e-3, name: Optional[str] = None):
        multiprocessing.Process.__init__(self)
        self.daemon = False
        self._exit = multiprocessing.Event()
        self._in_queue: multiprocessing.Queue[Any] = multiprocessing.Queue()
        self.timeout = timeout
        self.worker_name = name

    def put(self, data):
        logger.info(
            "put samples in worker-{}".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )
        self._in_queue.put(data)

    def run(self):
        logger.info(
            "start worker-{}".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )
        self._exit.clear()
        logger.info(
            "pre hook executed in worker-{}".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )
        self.pre()
        self.clear_queue()
        while not self._exit.is_set():
            try:
                data = self._in_queue.get(timeout=self.timeout)
                logger.info(
                    "consume samples in worker-{}".format(
                        self.worker_name if self.worker_name else os.getpid()
                    )
                )
                self.consume(data)
            except queue.Empty:
                # if queue is empty, loop to wait for next data until exiting
                pass
        logger.info(
            "post hook executed in worker-{}".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )
        self.post()
        self.clear_queue()
        logger.info(
            "worker{} exit".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )

    @abstractmethod
    def pre(self):
        pass

    @abstractmethod
    def consume(self, data):
        pass

    @abstractmethod
    def post(self):
        pass

    def stop(self):
        logger.info(
            "stop worker-{}".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )
        self._exit.set()

    def settimeout(self, timeout=0.01):
        self.timeout = timeout

    def clear_queue(self):
        logger.info(
            "clearing queue items in worker-{}".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )
        while True:
            try:
                self._in_queue.get(timeout=self.timeout)
            except queue.Empty:
                break
        logger.info(
            "all queue items in worker-{} are cleared".format(
                self.worker_name if self.worker_name else os.getpid()
            )
        )



import cProfile
import io
import pstats
from pstats import SortKey


def profile(func):
    def wrapper(*args, **kwargs):
        pr = cProfile.Profile()
        pr.enable()
        retval = func(*args, **kwargs)
        pr.disable()
        s = io.StringIO()
        sortby = SortKey.CUMULATIVE
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())
        return retval

    return wrapper



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/1/07
# License: MIT License
from typing import Optional, List, Tuple, Union
import warnings
import numpy as np
from numpy import ndarray
from scipy.linalg import solve
from scipy.signal import sosfiltfilt, cheby1, cheb1ord
from sklearn.base import BaseEstimator, TransformerMixin, clone


def robust_pattern(W: ndarray, Cx: ndarray, Cs: ndarray) -> ndarray:
    """Transform spatial filters to spatial patterns based on paper [1]_.

    Parameters
    ----------
    W : ndarray
        Spatial filters, shape (n_channels, n_filters).
    Cx : ndarray
        Covariance matrix of eeg data, shape (n_channels, n_channels).
    Cs : ndarray
        Covariance matrix of source data, shape (n_channels, n_channels).

    Returns
    -------
    A : ndarray
        Spatial patterns, shape (n_channels, n_patterns), each column is a spatial pattern.

    References
    ----------
    .. [1] Haufe, Stefan, et al. "On the interpretation of weight vectors of linear models in multivariate neuroimaging."
           Neuroimage 87 (2014): 96-110.
    """
    # use linalg.solve instead of inv, makes it more stable
    # see https://github.com/robintibor/fbcsp/blob/master/fbcsp/signalproc.py
    # and https://ww2.mathworks.cn/help/matlab/ref/mldivide.html
    A = solve(Cs.T, np.dot(Cx, W).T).T
    return A


class FilterBank(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        base_estimator: BaseEstimator,
        filterbank: List[ndarray],
        n_jobs: Optional[int] = None,
    ):
        self.base_estimator = base_estimator
        self.filterbank = filterbank
        self.n_jobs = n_jobs

    def fit(self, X: ndarray, y: Optional[ndarray] = None, **kwargs):
        self.estimators_ = [
            clone(self.base_estimator) for _ in range(len(self.filterbank))
        ]
        X = self.transform_filterbank(X)
        for i, est in enumerate(self.estimators_):
            est.fit(X[i], y, **kwargs)
        # def wrapper(est, X, y, kwargs):
        #     est.fit(X, y, **kwargs)
        #     return est
        # self.estimators_ = Parallel(n_jobs=self.n_jobs)(
        #     delayed(wrapper)(est, X[i], y, kwargs) for i, est in enumerate(self.estimators_))
        return self

    def transform(self, X: ndarray, **kwargs):
        X = self.transform_filterbank(X)
        feat = [est.transform(X[i], **kwargs) for i, est in enumerate(self.estimators_)]
        # def wrapper(est, X, kwargs):
        #     retval = est.transform(X, **kwargs)
        #     return retval
        # feat = Parallel(n_jobs=self.n_jobs)(
        #     delayed(wrapper)(est, X[i], kwargs) for i, est in enumerate(self.estimators_))
        feat = np.concatenate(feat, axis=-1)
        return feat

    def transform_filterbank(self, X: ndarray):
        Xs = np.stack([sosfiltfilt(sos, X, axis=-1) for sos in self.filterbank])
        return Xs


class FilterBankSSVEP(FilterBank):
    """Filter bank analysis for SSVEP."""

    def __init__(
        self,
        filterbank: List[ndarray],
        base_estimator: BaseEstimator,
        filterweights: Optional[ndarray] = None,
        n_jobs: Optional[int] = None,
    ):
        self.filterweights = filterweights
        super().__init__(base_estimator, filterbank, n_jobs=n_jobs)

    def transform(self, X: ndarray):  # type: ignore[override]
        features = super().transform(X)
        if self.filterweights is None:
            return features
        else:
            features = np.reshape(
                features, (features.shape[0], len(self.filterbank), -1)
            )
            return np.sum(
                features * self.filterweights[np.newaxis, :, np.newaxis], axis=1
            )


def generate_filterbank(
    passbands: List[Tuple[float, float]],
    stopbands: List[Tuple[float, float]],
    srate: int,
    order: Optional[int] = None,
    rp: float = 0.5,
):
    filterbank = []
    for wp, ws in zip(passbands, stopbands):
        if order is None:
            N, wn = cheb1ord(wp, ws, 3, 40, fs=srate)
            sos = cheby1(N, rp, wn, btype="bandpass", output="sos", fs=srate)
        else:
            sos = cheby1(order, rp, wp, btype="bandpass", output="sos", fs=srate)

        filterbank.append(sos)
    return filterbank


def generate_cca_references(
    freqs,
    srate,
    T,
    phases: Optional[Union[ndarray, int, float]] = None,
    n_harmonics: int = 1,
):
    if isinstance(freqs, int) or isinstance(freqs, float):
        freqs = [freqs]
    freqs = np.array(freqs)[:, np.newaxis]
    if phases is None:
        phases = 0
    if isinstance(phases, int) or isinstance(phases, float):
        phases = np.array([phases])
    phases = np.array(phases)[:, np.newaxis]
    t = np.linspace(0, T, int(T * srate))

    Yf = []
    for i in range(n_harmonics):
        Yf.append(
            np.stack(
                [
                    np.sin(2 * np.pi * (i + 1) * freqs * t + np.pi * phases),
                    np.cos(2 * np.pi * (i + 1) * freqs * t + np.pi * phases),
                ],
                axis=1,
            )
        )
    Yf = np.concatenate(Yf, axis=1)
    return Yf


def sign_flip(u, s, vh=None):
    """Flip signs of SVD or EIG using the method in paper [1]_.

    Parameters
    ----------
    u: ndarray
        left singular vectors, shape (M, K).
    s: ndarray
        singular values, shape (K,).
    vh: ndarray or None
        transpose of right singular vectors, shape (K, N).

    Returns
    -------
    u: ndarray
        corrected left singular vectors.
    s: ndarray
        singular values.
    vh: ndarray
        transpose of corrected right singular vectors.

    References
    ----------
    .. [1] https://www.sandia.gov/~tgkolda/pubs/pubfiles/SAND2007-6422.pdf
    """
    if vh is None:
        total_proj = np.sum(u * s, axis=0)
        signs = np.sign(total_proj)

        random_idx = signs == 0
        if np.any(random_idx):
            signs[random_idx] = 1
            warnings.warn(
                "The magnitude is close to zero, the sign will become arbitrary."
            )

        u = u * signs

        return u, s
    else:
        left_proj = np.sum(s[:, np.newaxis] * vh, axis=-1)
        right_proj = np.sum(u * s, axis=0)
        total_proj = left_proj + right_proj
        signs = np.sign(total_proj)

        random_idx = signs == 0
        if np.any(random_idx):
            signs[random_idx] = 1
            warnings.warn(
                "The magnitude is close to zero, the sign will become arbitrary."
            )

        u = u * signs
        vh = signs[:, np.newaxis] * vh

        return u, s, vh



# -*- coding: utf-8 -*-
# License: MIT License
"""
Amplifiers.

"""
import datetime
import socket
import struct
import threading
import time
from abc import abstractmethod
from collections import deque
from typing import List, Optional, Tuple, Dict

import numpy as np

from .logger import get_logger
from .workers import ProcessWorker

logger_amp = get_logger("amplifier")
logger_marker = get_logger("marker")


class RingBuffer(deque):
    """Online data RingBuffer.
    -author: Lichao Xu
    -Created on: 2021-04-01
    -update log:
        None
    Parameters
    ----------
        size: int,
            Size of the RingBuffer.
    """

    def __init__(self, size=1024):
        """Ring buffer object based on python deque data structure to store data.

        Parameters
        ----------
        size : int, optional
            maximum buffer size, by default 1024
        """
        super(RingBuffer, self).__init__(maxlen=size)
        self.max_size = size

    def isfull(self):
        """Whether current buffer is full or not.

        Returns
        ----------
        boolean
        """
        return len(self) == self.max_size

    def get_all(self):
        """Access all current buffer value.

        Returns
        ----------
        list
            the list of current buffer
        """
        return list(self)


class Marker(RingBuffer):
    """Intercept online data.
    -author: Lichao Xu
    -Created on: 2021-04-01
    -update log:
        2022-08-10 by Wei Zhao
    Parameters
    ----------
        interval: list,
            Time Window.
        srate: int,
            Amplifier setting sample rate.
        events: list,
            Event label.
    """

    def __init__(
        self, interval: list, srate: float, events: Optional[List[int]] = None
    ):
        self.events = events
        if events is not None:
            self.interval = [int(i * srate) for i in interval]
            self.latency = 0 if self.interval[1] <= 0 else self.interval[1]
            max_tlim = max(0, self.interval[0], self.interval[1])
            min_tlim = min(0, self.interval[0], self.interval[1])
            size = max_tlim - min_tlim
            if min_tlim >= 0:
                self.epoch_ind = [self.interval[0], self.interval[1]]
            else:
                self.epoch_ind = [
                    self.interval[0] - min_tlim,
                    self.interval[1] - min_tlim,
                ]
        else:
            # continous mode
            self.interval = [int(i * srate) for i in interval]
            self.latency = self.interval[1] - self.interval[0]
            size = self.latency
            self.epoch_ind = [0, size]

        self.countdowns: Dict[str, int] = {}
        self.is_rising = True
        super().__init__(size=size)

    def __call__(self, event: int):
        """Record label position.
        Parameters
        ----------
            event: int,
                Online real-time data tagging.
        """
        # add new countdown items
        if self.events is not None:
            event = int(event)
            if event != 0 and self.is_rising:
                if event in self.events:
                    # new_key = hashlib.md5(''.join(
                    #     [str(event), str(datetime.datetime.now())]).encode()).hexdigest()
                    new_key = "".join(
                        [
                            str(event),
                            datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S"),
                        ]
                    )
                    self.countdowns[new_key] = self.latency + 1
                    logger_marker.info("find new event {}".format(new_key))
                self.is_rising = False
            elif event == 0:
                self.is_rising = True
        else:
            if "fixed" not in self.countdowns:
                self.countdowns["fixed"] = self.latency

        drop_items = []
        # update countdowns
        for key, value in self.countdowns.items():
            value = value - 1
            if value == 0:
                drop_items.append(key)
                logger_marker.info("trigger epoch for event {}".format(key))
            self.countdowns[key] = value

        for key in drop_items:
            del self.countdowns[key]
        if drop_items and self.isfull():
            return True
        return False

    def get_epoch(self):
        """Fetch data from buffer."""
        data = super().get_all()
        return data[self.epoch_ind[0] : self.epoch_ind[1]]


class BaseAmplifier:
    """Base Ampifier class.
    -author: Lichao Xu
    -Created on: 2021-04-01
    -update log:
        2022-08-10 by Wei Zhao
    """

    def __init__(self):
        self._markers = {}
        self._workers = {}
        self._exit = threading.Event()

    @abstractmethod
    def recv(self):
        """the minimal recv data function, usually a package."""
        pass

    def start(self):
        """start the loop."""
        for work_name in self._workers:
            logger_amp.info("clear marker bufer")
            self._markers[work_name].clear()
        logger_amp.info("start the loop")
        self._t_loop = threading.Thread(target=self._inner_loop, name="main_loop")
        self._t_loop.start()

    def _inner_loop(self):
        """Inner loop in the threading."""
        self._exit.clear()
        logger_amp.info("enter the inner loop")
        while not self._exit.is_set():
            try:
                samples = self.recv()
                if samples:
                    self._detect_event(samples)
            except Exception:
                pass
        logger_amp.info("exit the inner loop")

    def stop(self):
        """stop the loop."""
        logger_amp.info("stop the loop")
        self._exit.set()
        logger_amp.info("waiting the child thread exit")
        self._t_loop.join()
        self.clear()

    def _detect_event(self, samples):
        """detect event label"""
        for work_name in self._workers:
            logger_amp.info("process worker-{}".format(work_name))
            marker = self._markers[work_name]
            worker = self._workers[work_name]
            for sample in samples:
                marker.append(sample)
                if marker(sample[-1]) and worker.is_alive():
                    worker.put(marker.get_epoch())

    def up_worker(self, name):
        logger_amp.info("up worker-{}".format(name))
        self._workers[name].start()

    def down_worker(self, name):
        logger_amp.info("down worker-{}".format(name))
        self._workers[name].stop()
        self._workers[name].clear_queue()

    def register_worker(self, name: str, worker: ProcessWorker, marker: Marker):
        logger_amp.info("register worker-{}".format(name))
        self._workers[name] = worker
        self._markers[name] = marker

    def unregister_worker(self, name: str):
        logger_amp.info("unregister worker-{}".format(name))
        del self._markers[name]
        del self._workers[name]

    def clear(self):
        logger_amp.info("clear all workers")
        worker_names = list(self._workers.keys())
        for name in worker_names:
            self._markers[name].clear()
            self.down_worker(name)
            self.unregister_worker(name)


class NeuroScan(BaseAmplifier):
    """An amplifier implementation for NeuroScan device.
    Intercept online data.
    -author: Lichao Xu
    -Created on: 2021-04-01
    -update log:
        2022-08-10 by Wei Zhao
    """

    _COMMANDS = {
        "stop_connect": b"CTRL\x00\x01\x00\x02\x00\x00\x00\x00",
        "start_acq": b"CTRL\x00\x02\x00\x01\x00\x00\x00\x00",
        "stop_acq": b"CTRL\x00\x02\x00\x02\x00\x00\x00\x00",
        "start_trans": b"CTRL\x00\x03\x00\x03\x00\x00\x00\x00",
        "stop_trans": b"CTRL\x00\x03\x00\x04\x00\x00\x00\x00",
        "show_ver": b"CTRL\x00\x01\x00\x01\x00\x00\x00\x00",
        "show_edf": b"CTRL\x00\x03\x00\x01\x00\x00\x00\x00",
        "start_imp": b"CTRL\x00\x02\x00\x03\x00\x00\x00\x00",
        "req_version": b"CTRL\x00\x01\x00\x01\x00\x00\x00\x00",
        "correct_dc": b"CTRL\x00\x02\x00\x05\x00\x00\x00\x00",
        "change_setup": b"CTRL\x00\x02\x00\x04\x00\x00\x00\x00",
    }

    def __init__(
        self,
        device_address: Tuple[str, int] = ("127.0.0.1", 4000),
        srate: float = 1000,
        num_chans: int = 68,
    ):
        super().__init__()
        self.device_address = device_address
        self.srate = srate
        self.num_chans = num_chans
        self.neuro_link = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # the size of a package in neuroscan data is srate/25*(num_chans+1)*4 bytes
        self.pkg_size = srate / 25 * (num_chans + 1) * 4
        self.timeout = 2 * 25 / self.srate

    def _unpack_header(self, b_header):
        ch_id = struct.unpack(">4s", b_header[:4])
        w_code = struct.unpack(">H", b_header[4:6])
        w_request = struct.unpack(">H", b_header[6:8])
        pkg_size = struct.unpack(">I", b_header[8:])
        return (ch_id[0].decode("utf-8"), w_code[0], w_request[0], pkg_size[0])

    def _unpack_data(self, num_chans, b_data):
        fmt = ">" + str((num_chans + 1) * 4) + "B"
        samples = (
            np.array(list(struct.iter_unpack(fmt, b_data)), dtype=np.uint8)
            .view(np.int32)
            .astype(np.float64)
        )
        samples[:, -1] = samples[:, -1] - 65280
        samples[:, :-1] = samples[:, :-1] * 0.0298 * 1e-6
        return samples.tolist()

    def _recv(self, num_bytes):
        fragments = []
        b_count = 0
        while b_count < num_bytes:
            try:
                chunk = self.neuro_link.recv(num_bytes - b_count)
            except socket.timeout as e:
                raise e
            b_count += len(chunk)
            fragments.append(chunk)

        b_data = b"".join(fragments)
        return b_data

    def recv(self):
        b_header = self._recv(12)
        header = self._unpack_header(b_header)
        samples = None
        if header[-1] != 0:
            b_data = self._recv(header[-1])
            samples = self._unpack_data(self.num_chans, b_data)
        return samples

    def send(self, message):
        self.neuro_link.sendall(message)

    def set_timeout(self, timeout):
        if self.neuro_link:
            self.neuro_link.settimeout(timeout)

    def command(self, method):
        if method == "connect":
            self.neuro_link.connect(self.device_address)
        elif method == "start_acquire":
            self.send(self._COMMANDS["start_acq"])
            self.set_timeout(None)
            self.recv()
            self.recv()
            self.set_timeout(self.timeout)
        elif method == "stop_acquire":
            self.set_timeout(None)
            self.send(self._COMMANDS["stop_acq"])
            self.recv()
            self.recv()
            self.set_timeout(self.timeout)
        elif method == "start_transport":
            self.send(self._COMMANDS["start_trans"])
            time.sleep(1e-2)
            self.start()
        elif method == "stop_transport":
            self.send(self._COMMANDS["stop_trans"])
            self.stop()
        elif method == "disconnect":
            self.send(self._COMMANDS["stop_connect"])
            if self.neuro_link:
                self.neuro_link.close()
                self.neuro_link = None

    def connect_tcp(self):
        self.neuro_link.connect(self.device_address)

    def start_acq(self):
        self.send(self._COMMANDS["start_acq"])
        self.set_timeout(None)
        self.recv()
        self.recv()
        self.set_timeout(self.timeout)

    def stop_acq(self):
        self.set_timeout(None)
        self.send(self._COMMANDS["stop_acq"])
        self.recv()
        self.recv()
        self.set_timeout(self.timeout)

    def start_trans(self):
        self.send(self._COMMANDS["start_trans"])
        time.sleep(1e-2)
        self.start()

    def stop_trans(self):
        self.send(self._COMMANDS["stop_trans"])
        self.stop()

    def close_connection(self):
        self.send(self._COMMANDS["stop_connect"])
        if self.neuro_link:
            self.neuro_link.close()
            self.neuro_link = None



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/01/07
# License: MIT License
"""
Tsinghua BCI Lab.
"""
import os
import zipfile
from typing import Union, Optional, Dict, List, cast
from pathlib import Path

import numpy as np
import py7zr
from mne import create_info
from mne.io import RawArray, Raw
from mne.channels import make_standard_montage
from .base import BaseDataset
from ..utils.download import mne_data_path
from ..utils.io import loadmat

# TSINGHUA_URL = 'http://bci.med.tsinghua.edu.cn/download.html'

# 403 error, though it still works
Wang2016_URL = "http://bci.med.tsinghua.edu.cn/upload/yijun/"
# Wang2016_URL = "ftp://sccn.ucsd.edu/pub/ssvep_benchmark_dataset/"
# Wang2016_URL = 'http://www.thubci.com/uploads/down/' # This may work
BETA_URL = "http://bci.med.tsinghua.edu.cn/upload/liubingchuan/"  # 403 error
# BETA_URL = 'https://figshare.com/articles/The_BETA_database/12264401'


class Wang2016(BaseDataset):
    """SSVEP dataset from Yijun Wang.

    This dataset gathered SSVEP-BCI recordings of 35 healthy subjects (17 females, aged 17-34 years, mean age: 22 years)
    focusing on 40 characters flickering at different frequencies (8-15.8 Hz with an interval of 0.2 Hz). For each subject,
    the experiment consisted of 6 blocks. Each block contained 40 trials corresponding to all 40 characters indicated in a
    random order. Each trial started with a visual cue (a red square) indicating a target stimulus. The cue appeared for
    0.5 s on the screen. Subjects were asked to shift their gaze to the target as soon as possible within the cue duration.
    Following the cue offset, all stimuli started to flicker on the screen concurrently and lasted 5 s. After stimulus offset,
    the screen was blank for 0.5 s before the next trial began, which allowed the subjects to have short breaks between
    consecutive trials. Each trial lasted a total of 6 s. To facilitate visual fixation, a red triangle appeared below the
    flickering target during the stimulation period. In each block, subjects were asked to avoid eye blinks during the
    stimulation period. To avoid visual fatigue, there was a rest for several minutes between two consecutive blocks.

    EEG data were acquired using a Synamps2 system (Neuroscan, Inc.) with a sampling rate of 1000 Hz. The amplifier frequency
    passband ranged from 0.15 Hz to 200 Hz. Sixty-four channels covered the whole scalp of the subject and were aligned
    according to the international 10-20 system. The ground was placed on midway between Fz and FPz. The reference was located
    on the vertex. Electrode impedances were kept below 10 KΩ. To remove the common power-line noise, a notch filter at 50 Hz
    was applied in data recording. Event triggers generated by the computer to the amplifier and recorded on an event channel
    synchronized to the EEG data.

    The continuous EEG data was segmented into 6 s epochs (500 ms pre-stimulus, 5.5 s post-stimulus onset). The epochs were
    subsequently downsampled to 250 Hz. Thus each trial consisted of 1500 time points. Finally, these data were stored as
    double-precision floating-point values in MATLAB and were named as subject indices (i.e., S01.mat, …, S35.mat). For each
    file, the data loaded in MATLAB generate a 4-D matrix named ‘data’ with dimensions of [64, 1500, 40, 6]. The four
    dimensions indicate ‘Electrode index’, ‘Time points’, ‘Target index’, and ‘Block index’. The electrode positions were
    saved in a ‘64-channels.loc’ file. Six trials were available for each SSVEP frequency. Frequency and phase values for
    the 40 target indices were saved in a ‘Freq_Phase.mat’ file.

    Information for all subjects was listed in a ‘Sub_info.txt’ file. For each subject, there are five factors including
    ‘Subject Index’, ‘Gender’, ‘Age’, ‘Handedness’, and ‘Group’. Subjects were divided into an ‘experienced’ group (eight
    subjects, S01-S08) and a ‘naive’ group (27 subjects, S09-S35) according to their experience in SSVEP-based BCIs.

    Frequency Table
    8    9   10   11   12   13   14   15
    8.2  9.2 10.2 11.2 12.2 13.2 14.2 15.2
    8.4  9.4 10.4 11.4 12.4 13.4 14.4 15.4
    8.6  9.6 10.6 11.6 12.6 13.6 14.6 15.6
    8.8  9.8 10.8 11.8 12.8 13.8 14.8 15.8

    Notes
    -----
    1. sub5 is not available from the download url.
    """

    _CHANNELS = [
        "FP1",
        "FPZ",
        "FP2",
        "AF3",
        "AF4",
        "F7",
        "F5",
        "F3",
        "F1",
        "FZ",
        "F2",
        "F4",
        "F6",
        "F8",
        "FT7",
        "FC5",
        "FC3",
        "FC1",
        "FCZ",
        "FC2",
        "FC4",
        "FC6",
        "FT8",
        "T7",
        "C5",
        "C3",
        "C1",
        "CZ",
        "C2",
        "C4",
        "C6",
        "T8",
        "TP7",
        "CP5",
        "CP3",
        "CP1",
        "CPZ",
        "CP2",
        "CP4",
        "CP6",
        "TP8",
        "P7",
        "P5",
        "P3",
        "P1",
        "PZ",
        "P2",
        "P4",
        "P6",
        "P8",
        "PO7",
        "PO5",
        "PO3",
        "POZ",
        "PO4",
        "PO6",
        "PO8",
        "O1",
        "OZ",
        "O2",
    ]

    _FREQS = [
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        8.2,
        9.2,
        10.2,
        11.2,
        12.2,
        13.2,
        14.2,
        15.2,
        8.4,
        9.4,
        10.4,
        11.4,
        12.4,
        13.4,
        14.4,
        15.4,
        8.6,
        9.6,
        10.6,
        11.6,
        12.6,
        13.6,
        14.6,
        15.6,
        8.8,
        9.8,
        10.8,
        11.8,
        12.8,
        13.8,
        14.8,
        15.8,
    ]

    _PHASES = [
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
    ]

    _EVENTS = {str(freq): (i + 1, (0, 5)) for i, freq in enumerate(_FREQS)}

    def __init__(self):
        super().__init__(
            dataset_code="wang2016",
            subjects=list(range(1, 36)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=250,
            paradigm="ssvep",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:
        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        subject = cast(int, subject)
        url = "{:s}S{:d}.mat.7z".format(Wang2016_URL, subject)
        file_dest = mne_data_path(
            url,
            "tsinghua",
            path=path,
            proxies=proxies,
            force_update=force_update,
            update_path=update_path,
        )

        if not os.path.exists(file_dest[:-3]):
            # decompression the data
            with py7zr.SevenZipFile(file_dest, "r") as archive:
                archive.extractall(path=Path(file_dest).parent)
        dests = [[file_dest[:-3]]]
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)
        raw_mat = loadmat(dests[0][0])
        epoch_data = raw_mat["data"] * 1e-6
        stim = np.zeros((1, *epoch_data.shape[1:]))
        # insert event label at stimulus-onset
        # 0.5s latency
        stim[0, 125] = np.tile(
            np.arange(1, 41)[:, np.newaxis], (1, epoch_data.shape[-1])
        )
        epoch_data = np.concatenate((epoch_data, stim), axis=0)
        data = np.transpose(epoch_data, (0, 3, 2, 1))

        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]
        ch_names = [ch_name.upper() for ch_name in self._CHANNELS]
        ch_names.insert(32, "M1")
        ch_names.insert(42, "M2")
        ch_names.insert(59, "CB1")
        ch_names = ch_names + ["CB2", "STI 014"]
        ch_types = ["eeg"] * 65
        ch_types[59] = "misc"
        ch_types[63] = "misc"
        ch_types[-1] = "stim"

        info = create_info(ch_names=ch_names, ch_types=ch_types, sfreq=self.srate)

        runs = dict()
        for i in range(data.shape[1]):
            raw = RawArray(
                data=np.reshape(data[:, i, ...], (data.shape[0], -1)), info=info
            )
            raw.set_montage(montage)
            runs["run_{:d}".format(i)] = raw

        sess = {"session_0": runs}
        return sess

    def get_freq(self, event: str):
        return self._FREQS[self._EVENTS[event][0] - 1]

    def get_phase(self, event: str):
        return self._PHASES[self._EVENTS[event][0] - 1]


class BETA(BaseDataset):
    """BETA SSVEP dataset [1]_.

    EEG data after preprocessing are store as a 4-way tensor, with a dimension of channel x time point x block x condition.
    Each trial comprises 0.5-s data before the event onset and 0.5-s data after the time window of 2 s or 3 s. For S1-S15,
    the time window is 2 s and the trial length is 3 s, whereas for S16-S70 the time window is 3 s and the trial length is
    4 s. Additional details about the channel and condition information can be found in the following supplementary
    information.

    Eight supplementary information is comprised of personal information, channel information, frequency and initial phase
    associated to each condition, SNR and sampling rate. The personal information contains age and gender of the subject.
    For the channel information, a location matrix (64 x 4) is provided, with the first column indicating channel index,
    the second column and third column indicating the degree and radius in polar coordinates, and the last column indicating
    channel name. The SNR information contains the mean narrow-band SNR and wide-band SNR matrix for each subject, calculated
    in (3) and (4), respectively. The initial phase is in radius.

    3-100Hz bandpass filtering (eegfilt), downsampled to 250 Hz

    References
    ----------
    .. [1] Liu B, Huang X, Wang Y, et al. BETA: A Large Benchmark Database Toward SSVEP-BCI Application[J]. Frontiers in
           neuroscience, 2020, 14: 627.
    """

    _CHANNELS = [
        "FP1",
        "FPZ",
        "FP2",
        "AF3",
        "AF4",
        "F7",
        "F5",
        "F3",
        "F1",
        "FZ",
        "F2",
        "F4",
        "F6",
        "F8",
        "FT7",
        "FC5",
        "FC3",
        "FC1",
        "FCZ",
        "FC2",
        "FC4",
        "FC6",
        "FT8",
        "T7",
        "C5",
        "C3",
        "C1",
        "CZ",
        "C2",
        "C4",
        "C6",
        "T8",
        "TP7",
        "CP5",
        "CP3",
        "CP1",
        "CPZ",
        "CP2",
        "CP4",
        "CP6",
        "TP8",
        "P7",
        "P5",
        "P3",
        "P1",
        "PZ",
        "P2",
        "P4",
        "P6",
        "P8",
        "PO7",
        "PO5",
        "PO3",
        "POZ",
        "PO4",
        "PO6",
        "PO8",
        "O1",
        "OZ",
        "O2",
    ]

    _FREQS = [
        8.6,
        8.8,
        9,
        9.2,
        9.4,
        9.6,
        9.8,
        10,
        10.2,
        10.4,
        10.6,
        10.8,
        11,
        11.2,
        11.4,
        11.6,
        11.8,
        12,
        12.2,
        12.4,
        12.6,
        12.8,
        13,
        13.2,
        13.4,
        13.6,
        13.8,
        14,
        14.2,
        14.4,
        14.6,
        14.8,
        15,
        15.2,
        15.4,
        15.6,
        15.8,
        8,
        8.2,
        8.4,
    ]
    _PHASES = [
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
        1.5,
        0,
        0.5,
        1,
    ]

    _EVENTS = {str(freq): (i + 1, (0, 2)) for i, freq in enumerate(_FREQS)}

    def __init__(self):
        super().__init__(
            dataset_code="beta",
            subjects=list(range(1, 71)),
            events=self._EVENTS,
            channels=self._CHANNELS,
            srate=250,
            paradigm="ssvep",
        )

    def data_path(
        self,
        subject: Union[str, int],
        path: Optional[Union[str, Path]] = None,
        force_update: bool = False,
        update_path: Optional[bool] = None,
        proxies: Optional[Dict[str, str]] = None,
        verbose: Optional[Union[bool, str, int]] = None,
    ) -> List[List[Union[str, Path]]]:
        if subject not in self.subjects:
            raise (ValueError("Invalid subject id"))

        subject = cast(int, subject)
        if subject < 11:
            url = "{:s}S1-S10.mat.zip".format(BETA_URL)
        elif subject < 21:
            url = "{:s}S11-S20.mat.zip".format(BETA_URL)
        elif subject < 31:
            url = "{:s}S21-S30.mat.zip".format(BETA_URL)
        elif subject < 41:
            url = "{:s}S31-S40.mat.zip".format(BETA_URL)
        elif subject < 51:
            url = "{:s}S41-S50.mat.zip".format(BETA_URL)
        elif subject < 61:
            url = "{:s}S51-S60.mat.zip".format(BETA_URL)
        else:
            url = "{:s}S61-S70.mat.zip".format(BETA_URL)

        file_dest = mne_data_path(
            url,
            "tsinghua",
            path=path,
            proxies=proxies,
            force_update=force_update,
            update_path=update_path,
        )

        parent_dir = Path(file_dest).parent

        if not os.path.exists(os.path.join(parent_dir, "S{:d}.mat".format(subject))):
            # decompression the data
            with zipfile.ZipFile(file_dest, "r") as archive:
                archive.extractall(path=parent_dir)
        dests: List[List[Union[str, Path]]] = [
            [os.path.join(parent_dir, "S{:d}.mat".format(subject))]
        ]
        return dests

    def _get_single_subject_data(
        self, subject: Union[str, int], verbose: Optional[Union[bool, str, int]] = None
    ) -> Dict[str, Dict[str, Raw]]:
        dests = self.data_path(subject)
        raw_mat = loadmat(dests[0][0])
        epoch_data = raw_mat["data"]["EEG"] * 1e-6
        stim = np.zeros((1, *epoch_data.shape[1:]))
        # 0.5s latency
        stim[0, 125] = np.tile(np.arange(1, 41), (epoch_data.shape[-2], 1))
        epoch_data = np.concatenate((epoch_data, stim), axis=0)
        data = np.transpose(epoch_data, (0, 3, 2, 1))

        montage = make_standard_montage("standard_1005")
        montage.rename_channels(
            {ch_name: ch_name.upper() for ch_name in montage.ch_names}
        )
        # montage.ch_names = [ch_name.upper() for ch_name in montage.ch_names]
        ch_names = [ch_name.upper() for ch_name in self._CHANNELS]
        ch_names.insert(32, "M1")
        ch_names.insert(42, "M2")
        ch_names.insert(59, "CB1")
        ch_names = ch_names + ["CB2", "STI 014"]
        ch_types = ["eeg"] * 65
        ch_types[59] = "misc"
        ch_types[63] = "misc"
        ch_types[-1] = "stim"

        info = create_info(ch_names=ch_names, ch_types=ch_types, sfreq=self.srate)

        runs = dict()
        for i in range(data.shape[-2]):
            raw = RawArray(
                data=np.reshape(data[..., i, :], (data.shape[0], -1)), info=info
            )
            raw.set_montage(montage)
            runs["run_{:d}".format(i)] = raw

        sess = {"session_0": runs}
        return sess

    def get_freq(self, event: str):
        return self._FREQS[self._EVENTS[event][0] - 1]

    def get_phase(self, event: str):
        return self._PHASES[self._EVENTS[event][0] - 1]



from .riemann import (
    logmap,
    expmap,
    geodesic,
    distance_riemann,
    mean_riemann,
    vectorize,
    unvectorize,
    tangent_space,
    untangent_space,
    mdrm_kernel,
    MDRM,
    FgMDRM,
    TSClassifier,
    FGDA,
    Alignment,
    RecursiveAlignment,
)
from .rpa import get_recenter, get_rescale, get_rotate, recenter, rescale, rotate



from scipy.fftpack import fft
from scipy import signal
import numpy as np
import matplotlib.pyplot as plt
import mne


class FrequencyAnalysis:
    def __init__(self, data, meta, event, srate, latency=0, channel="all"):
        """
        -author: Zhou hongzhan
        -Create on:2022-8-9
        -update log:
         2022-8-11 by Zhou hongzhan

         Args:
             1.data:EEG data (nTrials, nChannels, nTimes)
                 A matrix fullfilled with timepoint voltage
             2.meta:DataFrame
                 Concrete message of data,including subject ID,the events correspond to  specific trials,etc.
             3.event:String
                 Events needed to be extracted
             4.srate:Int
                 The sample rate of data
             5.latency:Float
                 The start timepoint of experiment (latency=0 indicate that the data recording begin with the stimuli performs)
                 default value=0
             6.channel:String
                 The wanted channel.if 'all',all channel will be extracted.default value = 'all'

        """

        sub_meta = meta[meta["event"] == event]
        event_id = sub_meta.index.to_numpy()
        self.data_length = np.round(data.shape[2] / srate)
        if channel == "all":
            self.data = data[event_id, :, :]
        else:
            self.data = data[event_id, channel, :]
        self.latency = latency
        self.fs = srate

    def stacking_average(self, data=[], _axis=0):
        """
        -author: Zhou hongzhan
        -Create on:2022-8-9
        -update log:
            2022-8-11 by Zhou hongzhan

        Args:
            data : np.array (nTrials, nChannels, nTimes)
                 EEG origin data. The default is [].
            _axis : int
                 The dimension need to be stacked. The default is 0.

        Returns:
            data_mean : np.array
                 The data after stacked.

        """
        if data == []:
            data = self.data
        data_mean = data
        data_mean = np.mean(data, axis=_axis)
        return data_mean

    def power_spectrum_periodogram(self, x):
        """
        -author: Zhou hongzhan & He Jiatong
        -Create on:2022-8-9
        -update log:
            2022-8-31 by Zhou hongzhan

        Args:
            x : np.array
               1D data.

        Returns:
            f : np.array
               An array of frequencies
            Pxx_den : np.array
               The amplitude array respectively correspond to frequency array

        """
        f, Pxx_den = signal.periodogram(x, self.fs, window="boxcar", scaling="spectrum")
        plt.plot(f, Pxx_den)
        plt.title("Power Spectral Density")
        plt.xlim([0, 60])
        plt.ylim([0, 0.5])
        plt.xlabel("frequency [Hz]")
        plt.ylabel("PSD [V**2]")
        plt.show()
        return f, Pxx_den

    def sum_y(self, x, y, x_inf, x_sup):
        """
        -author: Zhou hongzhan
        -Create on:2022-8-9
        -update log:
            2022-8-11 by Zhou hongzhan

        Args:
            x : np.array(1D)
               An array of frequencies
            y : np.array(1D,SAME TYPE WITH X)
               The amplitude array respectively correspond to frequency array
            x_inf : int
               Infimum of freq.
            x_sup : int
               Supremum of freq.

        Returns:
           np.mean(sum_A): int
               freq parameter,topomap procedure needed
        """
        sum_A = []
        for i, freq in enumerate(x):
            if freq <= x_sup and freq >= x_inf:
                sum_A.append(y[i])
        return np.mean(sum_A)

    def plot_topomap(self, data, ch_names, srate=-1, ch_types="eeg"):
        """
        -author: Zhou hongzhan & He Jiatong
        -Create on:2022-8-9
        -update log:
            2022-8-31 by Zhou hongzhan

        Args:
           data : np.array, 1D array
               eeg data. The default is [].
           ch_names : list
               interested channels
           srate : int
               sample rate. The default is -1.if set as default ,the initial sample ratio will be applied
           ch_types : string
               Type of channels,default value='eeg'
        """

        if srate == -1:
            srate = self.fs
        info = mne.create_info(ch_names=ch_names, sfreq=srate, ch_types=ch_types)
        evoked = mne.EvokedArray(data, info)
        evoked.set_montage("standard_1005")
        mne.viz.plot_topomap(evoked.data[:, 0], evoked.info, show=True)

    def signal_noise_ratio(self, data=[], srate=-1, T=[], channel=[]):
        """
        -author: Zhou hongzhan & He Jiatong
        -Create on:2022-8-9
        -update log:
            2022-8-31 by Zhou hongzhan

        Args:
            data : np.array, 1D array
               eeg data. The default is [].
            srate : int
               sample rate. The default is -1.if set as default ,the initial sample ratio will be applied
            T : int, ms
               the during time of data. The default is [].
            channel : string
               interested channels

        Returns:
            X1 : np.array
               frequency sequece.
            snr : np.array
               SNR sequence

        """
        if srate == -1:
            srate = self.fs
        num_fft = srate * T
        df = srate / num_fft
        n = np.arange(0, num_fft - 1, 1)
        fx = n * df
        fx = fx[None, :]
        Y = fft(data[channel, :], num_fft)
        Y = np.abs(Y) * 2 / num_fft
        Y = Y[None, :]
        X1 = fx[0, 0 : num_fft // 2]
        Y1 = Y[0, 0 : num_fft // 2]
        plt.plot(X1, Y1)
        plt.title("fft transform")
        plt.xlim(0, 60)
        plt.ylim(0, 2)
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Amplitude (μV)")
        plt.show()

        nn1 = np.linspace(
            start=5, stop=round(60 / df), num=round(60 / df), endpoint=False
        ).astype(int)
        snr = []
        for center_freq in nn1:
            if center_freq < round(60 / df):
                sum_denominator = []
                snr_nominator = []
                sum_denominator = np.mean(
                    Y1[center_freq - 5 : center_freq - 1]
                ) + np.mean(Y1[center_freq + 1 : center_freq + 5])
                snr_nominator = Y1[center_freq]
                SNR = 20 * np.log10(snr_nominator / sum_denominator)
                snr.append(SNR)
        plt.plot(X1[0 : round(60 / df)], snr)
        plt.title("Signal Noise Ratio")
        plt.xlim(5, 60)
        plt.ylim(-35, 10)
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Signal Noise Ratio (dB)")
        plt.show()
        return X1, snr



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/9/18
# License: MIT License
"""
SSCOR.
"""
from typing import Optional, List, Tuple
from functools import partial

import numpy as np
from scipy.linalg import eigh, cholesky, inv

from numpy import ndarray
from sklearn.base import BaseEstimator, TransformerMixin
from joblib import Parallel, delayed

from .base import robust_pattern, FilterBank


def sscor_kernel(
    X: ndarray, y: Optional[ndarray] = None, n_jobs: Optional[int] = None
) -> Tuple[ndarray, ndarray, ndarray]:
    """The kernel part in SSCOR algorithm based on paper[1]_., [2]_.

    Modified from https://github.com/mnakanishi/TRCA-SSVEP/blob/master/src/train_sscor.m

    Parameters
    ----------
    X : ndarray
        EEG data assuming removing mean, shape (n_trials, n_channels, n_samples)
    y : ndarray
        labels, shape (n_trials, ), not used here
    n_jobs: int, optional
        the number of jobs to use, default None

    Returns
    -------
    W: ndarray
        filters, shape (n_channels, n_filters)
    D: ndarray
        eigenvalues in descending order
    A: ndarray
        spatial patterns, shape (n_channels, n_filters)

    References
    ----------
    .. [1] Kumar G R K, Reddy M R. Designing a sum of squared correlations framework for enhancing SSVEP-based BCIs[J].
           IEEE Transactions on Neural Systems and Rehabilitation Engineering, 2019, 27(10): 2044-2050.
    .. [2] Kumar G R K, Reddy M R. Correction to “Designing a Sum of Squared Correlations Framework for Enhancing SSVEP
           Based BCIs”[J]. IEEE Transactions on Neural Systems and Rehabilitation Engineering, 2020, 28(4): 1044-1045.
    """
    X = np.copy(X)
    X = np.reshape(X, (-1, *X.shape[-2:]))
    X = X - np.mean(X, axis=-1, keepdims=True)
    mean_X = np.mean(X, axis=0)
    K1 = cholesky(mean_X @ mean_X.T)  # upper-triangular X=K.T@K
    iK1 = inv(K1)
    xC = mean_X @ np.transpose(X, axes=(0, 2, 1))
    C = X @ np.transpose(X, axes=(0, 2, 1))

    def target(iK1, xCi, Ci):
        Ki = cholesky(Ci)
        Gi = iK1.T @ xCi @ inv(Ki)
        return Gi.T @ Gi

    target = partial(target, iK1)
    G_T_G = np.sum(
        Parallel(n_jobs=n_jobs)(delayed(target)(xCi, Ci) for xCi, Ci in zip(xC, C)),
        axis=0,
    )
    D, W = eigh(G_T_G)
    ind = np.argsort(D)[::-1]
    D, W = D[ind], W[:, ind]
    W = iK1 @ W
    A = robust_pattern(W, G_T_G, W.T @ G_T_G @ W)
    return W, D, A


def sscor_feature(W: ndarray, X: ndarray, n_components: int = 1) -> ndarray:
    """Return sscor features.

    Modified from https://github.com/mnakanishi/TRCA-SSVEP/blob/master/src/test_sscor.m

    Parameters
    ----------
    W : ndarray
        spatial filters from csp_kernel, shape (n_channels, n_filters)
    X : ndarray
        eeg data, shape (n_trials, n_channels, n_samples)
    n_components : int, optional
        the first k components to use, usually even number, by default 1

    Returns
    -------
    ndarray
        features of shape (n_trials, n_components, n_samples)

    Raises
    ------
    ValueError
        n_components should less than half of the number of channels
    """
    W, X = np.copy(W), np.copy(X)
    max_components = W.shape[1]
    if n_components > max_components:
        raise ValueError("n_components should less than the number of channels")
    X = np.reshape(X, (-1, *X.shape[-2:]))
    X = X - np.mean(X, axis=-1, keepdims=True)
    features = np.matmul(W[:, :n_components].T, X)
    return features


class SSCOR(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        n_components: int = 1,
        transform_method: Optional[str] = None,
        ensemble: bool = False,
        n_jobs: Optional[int] = None,
    ):
        self.n_components = n_components
        self.transform_method = transform_method
        self.ensemble = ensemble
        self.n_jobs = n_jobs

    def fit(self, X: ndarray, y: ndarray):
        self.classes_ = np.unique(y)
        Ws, Ds, As = zip(
            *[
                sscor_kernel(X[y == label], n_jobs=self.n_jobs)
                for label in self.classes_
            ]
        )
        self.Ws_, self.Ds_, self.As_ = np.stack(Ws), np.stack(Ds), np.stack(As)

        self.templates_ = np.stack(
            [np.mean(X[y == label], axis=0) for label in self.classes_]
        )
        return self

    def transform(self, X: ndarray):
        n_components = self.n_components

        if self.transform_method is None:
            features = np.concatenate(
                [sscor_feature(W, X, n_components=n_components) for W in self.Ws_],
                axis=-2,
            )
            features = np.reshape(features, (features.shape[0], -1))
            return features
        elif self.transform_method == "corr":
            if self.ensemble:
                W = np.transpose(self.Ws_[..., :n_components], (1, 0, 2))
                W = np.reshape(W, (W.shape[0], -1))
                X = sscor_feature(W, X, n_components=W.shape[1])
                features = [
                    self._pearson_features(
                        X, sscor_feature(W, template, n_components=W.shape[1])
                    )
                    for template in self.templates_
                ]
            else:
                features = [
                    self._pearson_features(
                        sscor_feature(W, X, n_components=n_components),
                        sscor_feature(W, template, n_components=n_components),
                    )
                    for W, template in zip(self.Ws_, self.templates_)
                ]
            features = np.concatenate(features, axis=-1)
            return features
        else:
            raise ValueError("non-supported transform method")

    def _pearson_features(self, X: ndarray, templates: ndarray):
        X = np.reshape(X, (-1, *X.shape[-2:]))
        templates = np.reshape(templates, (-1, *templates.shape[-2:]))
        X = X - np.mean(X, axis=-1, keepdims=True)
        templates = templates - np.mean(templates, axis=-1, keepdims=True)
        X = np.reshape(X, (X.shape[0], -1))
        templates = np.reshape(templates, (templates.shape[0], -1))
        istd_X = 1 / np.std(X, axis=-1, keepdims=True)
        istd_templates = 1 / np.std(templates, axis=-1, keepdims=True)
        corr = (X @ templates.T) / (templates.shape[1] - 1)
        corr = istd_X * corr * istd_templates.T
        return corr


class FBSSCOR(FilterBank):
    """Filter Bank SSCOR method in paper [1]_., [2]_.

    filterbank and weights suggested in the paper.

    wp = [
        [6, 90], [14, 90], [22, 90], [30, 90], [38, 90], [46, 90], [54, 90], [62, 90], [70, 90], [78, 90]
    ]
    ws = [
        [4, 100], [10, 100], [16, 100], [24, 100], [32, 100], [40, 100], [48, 100], [56, 100], [64, 100], [72, 100]
    ]

    filterweights:
        np.arange(1, 11)**(-1.25) + 0.25

    References
    ----------
    .. [1] Kumar G R K, Reddy M R. Designing a sum of squared correlations framework for enhancing SSVEP-based BCIs[J].
           IEEE Transactions on Neural Systems and Rehabilitation Engineering, 2019, 27(10): 2044-2050.
    .. [2] Kumar G R K, Reddy M R. Correction to “Designing a Sum of Squared Correlations Framework for Enhancing SSVEP
           Based BCIs”[J]. IEEE Transactions on Neural Systems and Rehabilitation Engineering, 2020, 28(4): 1044-1045.
    """

    def __init__(
        self,
        n_components: int = 1,
        ensemble: bool = False,
        n_jobs: Optional[int] = None,
        filterbank: List[ndarray] = [],
        filterweights: Optional[ndarray] = None,
    ):
        self.n_components = n_components
        self.ensemble = ensemble
        self.n_jobs = n_jobs
        self.filterbank = filterbank
        self.filterweights = filterweights
        if filterweights is not None:
            if filterbank is None:
                self.filterweights = None
            else:
                if len(filterweights) != len(filterbank):
                    raise ValueError(
                        "the len of filterweights must be the same as that of filterbank"
                    )

        super().__init__(
            SSCOR(
                n_components=n_components,
                transform_method="corr",
                ensemble=ensemble,
                n_jobs=n_jobs,
            ),
            filterbank=filterbank,
        )

    def transform(self, X: ndarray):  # type: ignore[override]
        features = super().transform(X)
        if self.filterweights is None:
            return features
        else:
            features = np.reshape(
                features, (features.shape[0], len(self.filterbank), -1)
            )
            return np.sum(
                features * self.filterweights[np.newaxis, :, np.newaxis], axis=1
            )



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/1/23
# License: MIT License
"""
Riemannian Procrustes Analysis.
Modified from https://github.com/plcrodrigues/RPA
"""
from typing import Optional
from functools import partial
import numpy as np
from numpy import ndarray

import autograd.numpy as anp

try:
    from pymanopt.manifolds import Rotations
except Exception:
    from pymanopt.manifolds import SpecialOrthogonalGroup as Rotations
from pymanopt import Problem

try:
    from pymanopt.solvers import SteepestDescent
except Exception:
    from pymanopt.optimizers import SteepestDescent
from ..utils.covariance import covariances, sqrtm, invsqrtm, logm, powm
from .riemann import mean_riemann, distance_riemann


def get_recenter(
    X: ndarray, cov_method: str = "cov", mean_method: str = "riemann", n_jobs: int = 1
):
    X = np.reshape(X, (-1, *X.shape[-2:]))
    X = X - np.mean(X, axis=-1, keepdims=True)
    C = covariances(X, estimator=cov_method, n_jobs=n_jobs)
    if mean_method == "riemann":
        M = mean_riemann(C, n_jobs=n_jobs)
    elif mean_method == "euclid":
        M = np.mean(C, axis=0)
    iM12 = invsqrtm(M)
    return iM12


def recenter(X: ndarray, iM12: ndarray):
    X = np.reshape(X, (-1, *X.shape[-2:]))
    X = X - np.mean(X, axis=-1, keepdims=True)
    return iM12 @ X


def get_rescale(X: ndarray, cov_method: str = "cov", n_jobs: int = 1):
    X = np.reshape(X, (-1, *X.shape[-2:]))
    X = X - np.mean(X, axis=-1, keepdims=True)
    C = covariances(X, estimator=cov_method, n_jobs=n_jobs)
    M = mean_riemann(C, n_jobs=n_jobs)
    d = np.mean(np.square(distance_riemann(C, M, n_jobs=n_jobs)))
    scale = np.sqrt(1 / d)
    return M, scale


def rescale(
    X: ndarray, M: ndarray, scale: float, cov_method: str = "cov", n_jobs: int = 1
):
    X = np.reshape(X, (-1, *X.shape[-2:]))
    X = X - np.mean(X, axis=-1, keepdims=True)
    C = covariances(X, estimator=cov_method, n_jobs=n_jobs)
    iM12 = invsqrtm(M)
    M12 = sqrtm(M)
    A = iM12 @ C @ iM12
    B = M12 @ powm(A, (scale - 1) / 2) @ iM12
    X = B @ X
    return X


def _cost_euc(R: ndarray, Mt: ndarray, Ms: ndarray, weights: Optional[ndarray] = None):
    if weights is None:
        weights = anp.ones(len(Mt))

    cost = 0
    for i, a in enumerate(zip(Ms, Mt)):
        Msi, Mti = a
        Mti = anp.dot(R, anp.dot(Mti, R.T))
        cost += weights[i] * anp.square(anp.linalg.norm(Mti - Msi))
    # cost = anp.linalg.norm(Mt-Ms, ord='fro', axis=(-2, -1))
    return cost


def _cost_rie(R: ndarray, Mt: ndarray, Ms: ndarray, weights: Optional[ndarray] = None):
    if weights is None:
        weights = anp.ones(len(Mt))
    Mt = anp.matmul(R, anp.matmul(Mt, R.T))
    # distance_riemann not implemented in autograd, must provide egrad
    cost = anp.square(distance_riemann(Ms, Mt))
    return anp.dot(cost, weights)


def _egrad_rie(R: ndarray, Mt: ndarray, Ms: ndarray, weights: Optional[ndarray] = None):
    if weights is None:
        weights = anp.ones(len(Mt))
    # I dont't understand the code!!!
    iMt12 = invsqrtm(Mt)
    Ms12 = sqrtm(Ms)
    term_aux = anp.matmul(R, anp.matmul(Mt, R.T))
    term_aux = anp.matmul(iMt12, anp.matmul(term_aux, iMt12))
    g = 4 * np.matmul(np.matmul(iMt12, logm(term_aux)), np.matmul(Ms12, R))
    g = g * weights[:, np.newaxis, np.newaxis]
    return anp.sum(g, axis=0)


def _procruster_cost_function_euc(R, Mt, Ms):
    weights = anp.ones(len(Mt))

    c = []
    for Mti, Msi in zip(Mt, Ms):
        t1 = Msi
        t2 = anp.dot(R, anp.dot(Mti, R.T))
        ci = anp.linalg.norm(t1 - t2) ** 2
        c.append(ci)
    c = anp.array(c)

    return anp.dot(c, weights)


def _procruster_cost_function_rie(R, Mt, Ms):
    weights = anp.ones(len(Mt))

    c = []
    for Mti, Msi in zip(Mt, Ms):
        t1 = Msi
        t2 = anp.dot(R, anp.dot(Mti, R.T))
        ci = distance_riemann(t1, t2)[0] ** 2
        c.append(ci)
    c = anp.array(c)

    return anp.dot(c, weights)


def _procruster_egrad_function_rie(R, Mt, Ms):
    weights = anp.ones(len(Mt))

    g = []
    for Mti, Msi, wi in zip(Mt, Ms, weights):
        iMti12 = invsqrtm(Mti)
        Msi12 = sqrtm(Msi)
        term_aux = anp.dot(R, anp.dot(Msi, R.T))
        term_aux = anp.dot(iMti12, anp.dot(term_aux, iMti12))
        gi = 4 * anp.dot(anp.dot(iMti12, logm(term_aux)), anp.dot(Msi12, R))
        g.append(gi * wi)

    g = anp.sum(g, axis=0)

    return g


def _get_rotation_matrix(
    Mt: ndarray, Ms: ndarray, weights: Optional[ndarray] = None, metric: str = "euclid"
):
    Mt = Mt.reshape(-1, *Mt.shape[-2:])
    Ms = Ms.reshape(-1, *Ms.shape[-2:])

    n = Mt[0].shape[0]
    manifolds = Rotations(n)

    if metric == "euclid":
        # cost = partial(_cost_euc, Mt=Mt, Ms=Ms, weights=weights)
        cost = partial(_procruster_cost_function_euc, Mt=Mt, Ms=Ms)
        problem = Problem(manifold=manifolds, cost=cost, verbosity=0)
    elif metric == "riemann":
        # cost = partial(_cost_rie, Mt=Mt, Ms=Ms, weights=weights)
        # egrad = partial(_egrad_rie, Mt=Mt, Ms=Ms, weights=weights)
        cost = partial(_procruster_cost_function_rie, Mt=Mt, Ms=Ms)
        egrad = partial(_procruster_egrad_function_rie, Mt=Mt, Ms=Ms)
        problem = Problem(manifold=manifolds, cost=cost, egrad=egrad, verbosity=0)

    solver = SteepestDescent(mingradnorm=1e-3)

    Ropt = solver.solve(problem)

    return Ropt


def get_rotate(
    Xs: ndarray,
    ys: ndarray,
    Xt: ndarray,
    yt: ndarray,
    cov_method: str = "cov",
    metric: str = "euclid",
    n_jobs: int = 1,
):
    slabels = np.unique(ys)
    tlabels = np.unique(yt)
    Xs = np.reshape(Xs, (-1, *Xs.shape[-2:]))
    Xt = np.reshape(Xt, (-1, *Xt.shape[-2:]))
    Xs = Xs - np.mean(Xs, axis=-1, keepdims=True)
    Xt = Xt - np.mean(Xt, axis=-1, keepdims=True)
    Cs = covariances(Xs, estimator=cov_method, n_jobs=n_jobs)
    Ct = covariances(Xt, estimator=cov_method, n_jobs=n_jobs)

    Ms = np.stack([mean_riemann(Cs[ys == label]) for label in slabels])
    Mt = np.stack([mean_riemann(Ct[yt == label]) for label in tlabels])

    Ropt = _get_rotation_matrix(Mt, Ms, metric=metric)
    return Ropt


def rotate(Xt: ndarray, Ropt: ndarray):
    Xt = np.reshape(Xt, (-1, *Xt.shape[-2:]))
    Xt = Xt - np.mean(Xt, axis=-1, keepdims=True)
    return Ropt @ Xt



# -*- coding: utf-8 -*-
"""
Guney's network proposed in A Deep Neural Network for SSVEP-based Brain-Computer Interfaces.

Modified from https://github.com/osmanberke/Deep-SSVEP-BCI.git
"""
from collections import OrderedDict

import torch
import torch.nn as nn

from .base import (
    compute_same_pad2d,
    _narrow_normal_weight_zero_bias,
    compute_out_size,
    SkorchNet,
)


@SkorchNet
class GuneyNet(nn.Module):
    """
    Guney's network for decoding SSVEP.
    They used two stages to train the network.

    The first stage is with all training data in the dataset.
    lr: 1e-4, batch_size: 100, l2_regularization: 1e-3, epochs: 1000

    The second stage is a fine-tuning process with each subject's training data.
    lr: 1e-4, batch_size: full size, l2_regularization: 1e-3, epochs: 1000
    spatial_dropout=time1_dropout=0.6
    """

    def __init__(self, n_channels, n_samples, n_classes, n_bands):
        # super(GuneyNet, self).__init__()
        super().__init__()

        n_spatial_filters = 120
        spatial_dropout = 0.1
        time1_kernel = 2
        time1_stride = 2
        n_time1_filters = 120
        time1_dropout = 0.1
        time2_kernel = 10
        n_time2_filters = 120
        time2_dropout = 0.95

        self.n_channels = n_channels
        self.n_samples = n_samples
        self.n_classes = n_classes
        self.n_bands = n_bands

        self.model = nn.Sequential(
            OrderedDict(
                [
                    ("band_layer", nn.Conv2d(n_bands, 1, (1, 1), bias=False)),
                    ("spatial_layer", nn.Conv2d(1, n_spatial_filters, (n_channels, 1))),
                    ("spatial_dropout", nn.Dropout(spatial_dropout)),
                    (
                        "time1_layer",
                        nn.Conv2d(
                            n_spatial_filters,
                            n_time1_filters,
                            (1, time1_kernel),
                            stride=(1, time1_stride),
                        ),
                    ),
                    ("time1_dropout", nn.Dropout(time1_dropout)),
                    ("relu", nn.ReLU()),
                    (
                        "same_padding",
                        nn.ConstantPad2d(
                            compute_same_pad2d(
                                (
                                    1,
                                    compute_out_size(
                                        n_samples, time1_kernel, stride=time1_stride
                                    ),
                                ),
                                (1, time2_kernel),
                                stride=(1, 1),
                            ),
                            0,
                        ),
                    ),
                    (
                        "time2_layer",
                        nn.Conv2d(
                            n_time1_filters,
                            n_time2_filters,
                            (1, time2_kernel),
                            stride=(1, 1),
                        ),
                    ),
                    ("time2_dropout", nn.Dropout(time2_dropout)),
                    ("flatten", nn.Flatten()),
                    (
                        "fc_layer",
                        nn.Linear(
                            n_time2_filters
                            * compute_out_size(
                                n_samples, time1_kernel, stride=time1_stride
                            ),
                            n_classes,
                        ),
                    ),
                ]
            )
        )
        self._reset_parameters()

    @torch.no_grad()
    def _reset_parameters(self):
        _narrow_normal_weight_zero_bias(self)
        nn.init.ones_(self.model[0].weight)
        # MATLAB uses xavier_uniform_ with varaiance 2/(input+output)
        # perhaps this is a mistake in Help document
        nn.init.xavier_normal_(self.model[-1].weight, gain=1)

    def forward(self, X):
        # X: (n_batch, n_bands, n_channels, n_samples)
        out = self.model(X)
        return out



# -*- coding: utf-8 -*-
#
# Authors: Swolf <swolfforever@gmail.com>
# Date: 2021/7/06
# License: MIT License
"""
ShallowFBCSP.
Modified from https://github.com/braindecode/braindecode/blob/master/braindecode/models/shallow_fbcsp.py

"""

from collections import OrderedDict

import torch
import torch.nn as nn
from torch import Tensor
from .base import SkorchNet


class Square(nn.Module):
    def __init__(self):
        super(Square, self).__init__()

    def forward(self, X):
        return torch.square(X)


class SafeLog(nn.Module):
    def __init__(self, eps=1e-6):
        super(SafeLog, self).__init__()
        self.eps = eps

    def forward(self, X):
        return torch.log(torch.clamp(X, min=self.eps))


@SkorchNet
class ShallowNet(nn.Module):
    def __init__(self, n_channels: int, n_samples: int, n_classes: int):
        # super(ShallowNet, self).__init__()
        super().__init__()

        n_time_filters = 40
        time_kernel = 25
        n_space_filters = 40
        pool_kernel = 75
        pool_stride = 15
        dropout_rate = 0.5

        # temporal convolution
        self.step1 = nn.Sequential(
            OrderedDict(
                [
                    (
                        "time_conv",
                        nn.Conv2d(
                            1,
                            n_time_filters,
                            (1, time_kernel),
                            stride=1,
                            padding=0,
                            bias=True,
                        ),
                    )
                ]
            )
        )

        # spatial convolution
        self.step2 = nn.Sequential(
            OrderedDict(
                [
                    (
                        "space_conv",
                        nn.Conv2d(
                            n_time_filters,
                            n_space_filters,
                            (n_channels, 1),
                            stride=1,
                            padding=0,
                            bias=False,
                        ),
                    ),
                    ("bn", nn.BatchNorm2d(n_space_filters)),
                ]
            )
        )

        # pooling
        self.step3 = nn.Sequential(
            OrderedDict(
                [
                    ("square", Square()),
                    (
                        "avg_pool",
                        nn.AvgPool2d((1, pool_kernel), stride=(1, pool_stride)),
                    ),
                    ("safe_log", SafeLog()),
                    ("drop", nn.Dropout(p=dropout_rate)),
                    ("flatten", nn.Flatten()),
                ]
            )
        )

        with torch.no_grad():
            fake_input = torch.zeros((1, 1, n_channels, n_samples))
            fake_output = self.step3(self.step2(self.step1(fake_input)))
            middle_size = fake_output.shape[-1]

        self.fc_layer = nn.Linear(middle_size, n_classes, bias=True)
        self.model = nn.Sequential(self.step1, self.step2, self.step3, self.fc_layer)

        self._reset_parameters()

    @torch.no_grad()
    def _reset_parameters(self):
        nn.init.xavier_uniform_(self.step1.time_conv.weight, gain=1)
        nn.init.constant_(self.step1.time_conv.bias, 0)
        nn.init.xavier_uniform_(self.step2.space_conv.weight, gain=1)
        nn.init.constant_(self.step2.bn.weight, 1)
        nn.init.constant_(self.step2.bn.bias, 0)
        nn.init.xavier_uniform_(self.fc_layer.weight, gain=1)
        nn.init.constant_(self.fc_layer.bias, 0)

    def forward(self, X: Tensor, **kwargs):
        X = X.unsqueeze(1)
        out = self.model(X)
        return out



# -*- coding: utf-8 -*-
import datetime
import gc
from collections import OrderedDict
from functools import partial
import numpy as np
from psychopy import core, visual, event, logging

from .utils import _check_array_like, _clean_dict


class Experiment:
    """Paradigm start screen.
    -author: Lichao Xu
    -Created on: 2020-07-30
    -update log:
        2022-08-10 by Wei Zhao
    Parameters
    ----------
        win_size: ndarray,
            Size of the window in pixels [x, y].
        screen_id: int,
            The id of screen.
        is_fullscr: bool,
             Create a window in 'full-screen' mode.
        monitor:
             The monitor to be used during the experiment.
        bg_color_warm: ndarray,
            The start screen color.
        record_frames: bool,
            Record time elapsed per frame.
        disable_gc: bool,
            Garbage collector interface.
        process_priority: str
            The task priority.
        use_fbo:
            When drawing multiple windows, the FBO of a window can be switched.
    """

    def __init__(
        self,
        win_size=(800, 600),
        screen_id=0,
        is_fullscr=False,
        monitor=None,
        bg_color_warm=np.array([0, 0, 0]),
        record_frames=True,
        disable_gc=False,
        process_priority="normal",
        use_fbo=False,
    ):
        # global keys to exit experiment
        # only works in pyglet backend,the num lock key should be released first
        if not _check_array_like(win_size, 2):
            raise ValueError("win_size should be a 2 elements array-like object.")
        self.win_size = win_size

        if not isinstance(screen_id, int):
            raise ValueError("screen_id should be an int object")
        self.screen_id = screen_id

        if not _check_array_like(bg_color_warm, 3):
            raise ValueError("bg_color should be 3 elements array-like object.")
        self.bg_color_warm = bg_color_warm

        self.is_fullscr = is_fullscr
        self.monitor = monitor
        self.record_frames = record_frames

        # windows
        self.current_win = None

        # stimuli
        self.cache_stims = {}

        # paradigms
        self.paradigms = OrderedDict()
        self.current_paradigm = None

        # high performance twicking
        visual.useFBO = use_fbo
        if process_priority == "normal":
            pass
        elif process_priority == "high":
            core.rush(True)
        elif process_priority == "realtime":
            # Only makes a diff compared to 'high' on Windows.
            core.rush(True, realtime=True)
        else:
            print(
                "Invalid process priority:",
                process_priority,
                "Process running at normal.",
            )
            process_priority = "normal"

        if disable_gc:
            gc.disable()

    def initEvent(self):
        """Init operations before run."""
        self.global_clock = core.Clock()
        logging.setDefaultClock(self.global_clock)
        logging.console.setLevel(logging.WARNING)
        self.log_file = logging.LogFile(
            "logLastRun.log", filemode="w", level=logging.DATA
        )

        logging.warning(
            "============start experiment at {}============".format(
                datetime.datetime.now()
            )
        )
        event.clearEvents()
        event.globalKeys.add(key="escape", func=self.closeEvent)

    def closeEvent(self):
        """Close operation after run."""
        logging.warning(
            "============end Experiemnt at {}============".format(
                datetime.datetime.now()
            )
        )
        # fixed sys.meta_path error
        _clean_dict(self.cache_stims)
        # restore gamma map
        core.quit()

    def register_paradigm(self, name, func, *args, **kwargs):
        # fixed supplied arguments
        self.paradigms[name] = partial(func, *args, **kwargs)

    def unregister_paradigm(self, name):
        # clean stims
        self.cache_stims[name] = None
        del self.cache_stims[name]

        # clean paradigms
        self.paradigms[name] = None
        del self.paradigms[name]

    def get_window(self):
        """
        -update log:
            2022-08-10 by Wei Zhao
        """
        if not self.current_win:
            self.current_win = visual.Window(
                # the only-one option in psychopy, pygame is deprecated and glfw has lots of bugs
                winType="pyglet",
                units="pix",  # default pixel unit in this framework
                allowGUI=False,
                # Here are timing related options
                waitBlanking=False,  # much faster
                useFBO=False,
                checkTiming=True,
                numSamples=2,
                # user specified options
                size=self.win_size,
                screen=self.screen_id,
                monitor=self.monitor,
                fullscr=self.is_fullscr,
                color=self.bg_color_warm,
            )
            self.current_win.flip()
        return self.current_win

    def warmup(self, strict=True):
        win = self.get_window()
        fps = win.getActualFrameRate(
            nIdentical=10, nMaxFrames=100, nWarmUpFrames=10, threshold=1
        )  # keep int refresh rate
        if strict and fps is None:
            raise ValueError(
                "Can't get stable refresh rate. Close unnecessary programs or buy a better graphic cards."
            )

        self.fps = (
            int(np.rint(fps)) if fps is not None else int(1 / win.monitorFramePeriod)
        )
        self.frame_period = 1 / self.fps
        logging.warning(
            "Current screen refresh rate {}Hz and frame period {:.2f}ms".format(
                self.fps, self.frame_period * 1000
            )
        )
        win.recordFrameIntervals = self.record_frames
        win.refreshThreshold = 1 / self.fps + 0.002
        win.setMouseVisible(False)

    def update_startup(self):
        win = self.get_window()
        stims = self.cache_stims.setdefault("startup", OrderedDict())

        # check cache stims
        if "expguide_textstim" not in stims:
            stims["expguide_textstim"] = visual.TextStim(
                win,
                text="Welcome to the BCI world!\nPress Enter to select one of the following paradigms\nPress q to quit\n"
                "You can press esc to leave the program at any time!",
                units="height",
                pos=[0, 0.3],
                height=0.04,
                color="#ff944d",
                bold=False,
                alignText="center",
                anchorHoriz="center",
                wrapWidth=2,
            )

        # remove useless paradigms
        excludes = ["expguide_textstim"]
        stims = _clean_dict(stims, list(self.paradigms.keys()) + excludes)

        # update paradigm parameters
        names = list(self.paradigms.keys())
        if names and self.current_paradigm is None:
            self.current_paradigm = names[0]

        for i, name in enumerate(names):
            if name not in stims:
                stims[name] = visual.TextStim(
                    win,
                    text=name,
                    units="height",
                    pos=[0, -i * 0.03],
                    height=0.04,
                    color="#cccccc",
                    alignText="center",
                    anchorHoriz="center",
                    wrapWidth=1,
                )
            stims[name].setPos([0, -0.1 - i * 0.05])
            if name == self.current_paradigm:
                stims[name].setColor("#ff944d")
            else:
                stims[name].setColor("#cccccc")

        # draw all of them according to insert order
        for stim_name in stims:
            stims[stim_name].draw()

    def run(self):
        """Run the main loop."""
        self.initEvent()
        self.warmup()
        win = self.get_window()

        if self.record_frames:
            fps_textstim = visual.TextStim(
                win,
                text="",
                units="norm",
                pos=[-0.95, 0.95],
                height=0.03,
                color="#f2f2f2",
                alignText="left",
                anchorHoriz="left",
            )

        trialClock = core.Clock()
        t = lastFPSupdate = 0

        pindex = 0
        # capture runtime errors
        try:
            while True:
                t = trialClock.getTime()
                keys = event.getKeys(keyList=["q", "up", "down", "return"])

                # exit program
                if "q" in keys:
                    break

                # select paradigm
                names = list(self.paradigms.keys())
                if names:
                    if "up" in keys:
                        pindex -= 1
                        pindex = pindex % len(names)
                    elif "down" in keys:
                        pindex += 1
                        pindex = pindex % len(names)
                    self.current_paradigm = names[pindex]

                if "return" in keys:
                    old_color = win.color
                    logging.warning("Start paradigm {}".format(self.current_paradigm))
                    self.paradigms[self.current_paradigm](win=win)
                    logging.warning("Finish paradigm {}".format(self.current_paradigm))
                    win.color = old_color

                # main interface
                self.update_startup()

                if self.record_frames:
                    if t - lastFPSupdate > 1:
                        fps_textstim.text = "%i fps" % win.fps()
                        lastFPSupdate += 1
                    fps_textstim.draw()

                win.flip()

        except Exception as e:
            print("Error Info:", e)
            raise e
        finally:
            if self.record_frames:
                win.saveFrameIntervals("logLastFrameIntervals.log")
            win.close()
            self.closeEvent()
